"""
This module provides io for yaml files (.yaml extension)
See iotools.yamlio.help_yaml() for more info
"""

from iotools._help_utils import _help_default
from iotools._generic_open import _generic_open

try:
    import yaml
except ModuleNotFoundError:
    from iotools.settings import settings
    from iotools._missing_module_helper import _EmptyModule, _error_msg_one
    if settings["SHOW_IMPORT_WARNINGS"]:
        print("Warning:", _error_msg_one("iotools.yamlio", "PyYAML"))
    yaml = _EmptyModule("PyYAML")


def decode_yaml(data, **kwargs):
    """
    Decode yaml data from str
    Backend used: yaml.safe_load (PyYAML package)

    Args:
        data (str): yaml data to decode
        **kwargs: same as in 'yaml.safe_load'

    Returns:
        dict: yaml data decoded
    """
    if isinstance(data, bytes):
        data = data.decode("utf-8")
    return yaml.safe_load(data, **kwargs)


def encode_yaml(data, **kwargs):
    """
    Encode yaml data to str
    Backend used: yaml.safe_dump (PyYAML package)

    Args:
        data (dict): yaml data to encode
        **kwargs: same as in 'yaml.safe_dump'

    Returns:
        str: yaml data encoded
    """
    return yaml.safe_dump(data, **kwargs)


def read_yaml(openfile, **kwargs):
    """
    Read a open yaml file
    Backend used: yaml.safe_load (PyYAML package)

    Args:
        openfile (file-like): file to read, must have been opened
        **kwargs: same as in 'yaml.safe_load'

    Returns:
        dict: yaml data read from the file provided
    """
    return yaml.safe_load(openfile, **kwargs)


def write_yaml(openfile, data, **kwargs):
    """
    Write in a open yaml file
    Backend used: yaml.safe_dump (PyYAML package)

    Args:
        openfile (file-like): file to write, must have been opened
        data (dict): yaml data to save
        **kwargs: same as in 'yaml.safe_dump'
    """
    return yaml.safe_dump(data, openfile, **kwargs)


def load_yaml(filename, fs=None, **kwargs):
    """
    Load a yaml file
    Backend used: yaml.safe_load (PyYAML package)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        **kwargs: same as in 'yaml.safe_load'

    Returns:
        dict: loaded yaml data
    """
    with _generic_open(filename, 'r', fs=fs) as f:
        data = read_yaml(f, **kwargs)
    return data


def save_yaml(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a yaml file
    Backend used: yaml.safe_dump (PyYAML package)

    Args:
        filename (str): file name to save data to
        data (dict): yaml data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'yaml.safe_dump'
    """
    with _generic_open(filename, 'w', fs=fs, makedirs=makedirs) as f:
        write_yaml(f, data, **kwargs)


def help_yaml():
    """
    Print help for yaml io
    """
    _help_default("yaml")




import os

from fsspec.implementations.local import LocalFileSystem

localfs = LocalFileSystem()


def check_fs(fs):
    if (fs is None) or (fs is os):
        return localfs
    return fs


def is_localfs(fs):
    return (fs is None) or (fs is os) or isinstance(fs, LocalFileSystem)



"""
This module provides io for csv files (.csv extension)
See iotools.csvio.help_csv() for more info
"""

import csv

from iotools._help_utils import _help_default, _help_multi_packages
from iotools._generic_open import _generic_open
from iotools._default_decode_encode import _default_decode_str, _default_encode_str


try:
    import pandas as pd
except ModuleNotFoundError:
    from iotools._missing_module_helper import _EmptyModule
    pd = _EmptyModule("pandas")


# +-----------+
# | Using csv |
# +-----------+


def decode_csv_using_csv(data, **kwargs):
    """
    Decode csv data from str using csv library
    Backend used: csv.reader (csv package)

    Args:
        data (str): csv data to decode
        **kwargs: same as in 'csv.reader'

    Returns:
        list: csv data decoded
    """
    if isinstance(data, bytes):
        data = data.decode("utf-8")
    return _default_decode_str(read_csv_using_csv, data, **kwargs)


def encode_csv_using_csv(data, **kwargs):
    """
    Encode csv data to str using csv library
    Backend used: csv.writer (csv package)

    Args:
        data (list[list]): csv data to encode
        **kwargs: same as in 'csv.writer'

    Returns:
        str: csv data encoded
    """
    return _default_encode_str(write_csv_using_csv, data, **kwargs)


def read_csv_using_csv(openfile, **kwargs):
    """
    Read a open csv file using csv library
    Backend used: csv.reader (csv package)

    Args:
        openfile (file-like): file to read, must have been opened
        **kwargs: same as in 'csv.reader'

    Returns:
        list[list]: csv data read from the file provided
    """
    reader = csv.reader(openfile, **kwargs)
    return list(reader)


def write_csv_using_csv(openfile, data, **kwargs):
    """
    Write in a open csv file using csv library
    Backend used: csv.writer (csv package)

    Args:
        openfile (file-like): file to write, must have been opened
        data (list[list]): csv data to save
        **kwargs: same as in 'csv.writer'
    """
    writer = csv.writer(openfile, **kwargs)
    for row in data:
        writer.writerow(row)


def load_csv_using_csv(filename, fs=None, **kwargs):
    """
    Load a csv file using csv library
    Backend used: csv.reader (csv package)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        **kwargs: same as in 'csv.reader'

    Returns:
        list[list]: loaded csv data
    """
    with _generic_open(filename, 'r', fs=fs) as f:
        data = read_csv_using_csv(f, **kwargs)
    return data


def save_csv_using_csv(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a csv file using csv library
    Backend used: csv.writer (csv package)

    Args:
        filename (str): file name to save data to
        data (list[list]): csv data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'csv.writer'
    """
    with _generic_open(filename, 'w', fs=fs, makedirs=makedirs) as f:
        write_csv_using_csv(f, data, **kwargs)


def help_csv_using_csv():
    """
    Print help for csv io using csv library
    """
    _help_default("csv_using_csv")


# +--------------+
# | Using pandas |
# +--------------+


def decode_csv_using_pandas(data, **kwargs):
    """
    Decode csv data from str using pandas library
    Backend used: pandas.read_csv (pandas package)

    Args:
        data (str): csv data to decode
        **kwargs: same as in 'pandas.read_csv'

    Returns:
        pandas.DataFrame: csv data decoded
    """
    if isinstance(data, bytes):
        data = data.decode("utf-8")
    return _default_decode_str(read_csv_using_pandas, data, **kwargs)


def encode_csv_using_pandas(data, **kwargs):
    """
    Encode csv data to str using pandas library
    Backend used: pandas.DataFrame.to_csv (pandas package)

    Args:
        data (pandas.DataFrame): csv data to encode
        **kwargs: same as in 'pandas.DataFrame.to_csv'

    Returns:
        str: csv data encoded
    """
    return _default_encode_str(write_csv_using_pandas, data, **kwargs)


def read_csv_using_pandas(openfile, **kwargs):
    """
    Read a open csv file using pandas library
    Backend used: pandas.read_csv (pandas package)

    Args:
        openfile (file-like): file to read, must have been opened
        **kwargs: same as in 'pandas.read_csv'

    Returns:
        pandas.DataFrame: csv data read from the file provided
    """
    return pd.read_csv(openfile, **kwargs)


def write_csv_using_pandas(openfile, data, **kwargs):
    """
    Write in a open csv file using pandas library
    Backend used: pandas.DataFrame.to_csv (pandas package)

    Args:
        openfile (file-like): file to write, must have been opened
        data (pandas.DataFrame): csv data to save
        **kwargs: same as in 'pandas.DataFrame.to_csv'
    """
    pd.DataFrame(data).to_csv(openfile, **kwargs)


def load_csv_using_pandas(filename, fs=None, **kwargs):
    """
    Load a csv file using pandas library
    Backend used: pandas.read_csv (pandas package)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        **kwargs: same as in 'pandas.read_csv'

    Returns:
        pandas.DataFrame: loaded csv data
    """
    with _generic_open(filename, 'r', fs=fs) as f:
        data = read_csv_using_pandas(f, **kwargs)
    return data


def save_csv_using_pandas(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a csv file using pandas library
    Backend used: pandas.DataFrame.to_csv (pandas package)

    Args:
        filename (str): file name to save data to
        data (pandas.DataFrame): csv data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'pandas.DataFrame.to_csv'
    """
    with _generic_open(filename, 'w', fs=fs, makedirs=makedirs) as f:
        write_csv_using_pandas(f, data, **kwargs)


def help_csv_using_pandas():
    """
    Print help for csv io using pandas library
    """
    _help_default("csv_using_pandas")


# +------------------+
# | Default behavior |
# +------------------+


def decode_csv(data, **kwargs):
    return decode_csv_using_csv(data, **kwargs)


def encode_csv(data, **kwargs):
    return encode_csv_using_csv(data, **kwargs)


def read_csv(openfile, **kwargs):
    return read_csv_using_csv(openfile, **kwargs)


def write_csv(openfile, data, **kwargs):
    return write_csv_using_csv(openfile, data, **kwargs)


def load_csv(filename, fs=None, **kwargs):
    return load_csv_using_csv(filename, fs=fs, **kwargs)


def save_csv(filename, data, fs=None, makedirs=False, **kwargs):
    return save_csv_using_csv(filename, data, fs=fs, makedirs=makedirs, **kwargs)


decode_csv.__doc__ = decode_csv_using_csv.__doc__.replace(" using csv library", "")
encode_csv.__doc__ = encode_csv_using_csv.__doc__.replace(" using csv library", "")
read_csv.__doc__ = read_csv_using_csv.__doc__.replace(" using csv library", "")
write_csv.__doc__ = write_csv_using_csv.__doc__.replace(" using csv library", "")
load_csv.__doc__ = load_csv_using_csv.__doc__.replace(" using csv library", "")
save_csv.__doc__ = save_csv_using_csv.__doc__.replace(" using csv library", "")


def help_csv():
    """
    Print help for csv io using several libraries
    """
    _help_multi_packages("csv", "csv, pandas")



"""
This module provides io for xml files (.xml extension)
See iotools.xmlio.help_xml() for more info
"""

from iotools._help_utils import _help_default
from iotools._generic_open import _generic_open

try:
    import xmltodict
except ModuleNotFoundError:
    from iotools.settings import settings
    from iotools._missing_module_helper import _EmptyModule, _error_msg_one
    if settings["SHOW_IMPORT_WARNINGS"]:
        print("Warning:", _error_msg_one("iotools.xmlio", "xmltodict"))
    xmltodict = _EmptyModule("xmltodict")


def decode_xml(data, **kwargs):
    """
    Decode xml data from str
    Backend used: xmltodict.parse (xmltodict package)

    Args:
        data (str): xml data to decode
        **kwargs: same as in 'xmltodict.parse'

    Returns:
        OrderedDict: xml data decoded
    """
    if isinstance(data, bytes):
        data = data.decode("utf-8")
    return xmltodict.parse(data, **kwargs)


def encode_xml(data, **kwargs):
    """
    Encode xml data to str
    Backend used: xmltodict.unparse (xmltodict package)

    Args:
        data (OrderedDict | dict): xml data to encode
        **kwargs: same as in 'xmltodict.unparse'

    Returns:
        str: xml data encoded
    """
    return xmltodict.unparse(data, **kwargs)


def read_xml(openfile, **kwargs):
    """
    Read an open xml file
    Backend used: xmltodict.parse (xmltodict package)

    Args:
        openfile (file-like): file to read, must have been opened
        **kwargs: same as in 'xmltodict.parse'

    Returns:
        OrderedDict: xml data read from the file provided
    """
    data = openfile.read()
    return xmltodict.parse(data, **kwargs)


def write_xml(openfile, data, **kwargs):
    """
    Write in an open xml file
    Backend used: xmltodict.unparse (xmltodict package)

    Args:
        openfile (file-like): file to write, must have been opened
        data (OrderedDict | dict): xml data to save
        **kwargs: same as in 'xmltodict.unparse'
    """
    data = xmltodict.unparse(data, **kwargs)
    openfile.write(data)


def load_xml(filename, fs=None, **kwargs):
    """
    Load a xml file
    Backend used: xmltodict.parse (xmltodict package)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        **kwargs: same as in 'xmltodict.parse'

    Returns:
        OrderedDict: xml data read from the file provided
    """
    with _generic_open(filename, 'r', fs=fs) as f:
        data = read_xml(f, **kwargs)
    return data


def save_xml(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a xml file
    Backend used: xmltodict.unparse (xmltodict package)

    Args:
        filename (str): file name to save data to
        data (OrderedDict | dict): xml data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'xmltodict.unparse'
    """
    with _generic_open(filename, 'w', fs=fs, makedirs=makedirs) as f:
        write_xml(f, data, **kwargs)


def help_xml():
    """
    Print help for xml io
    """
    _help_default("xml")



"""
The module helps printing a comprehensible message for users who
did not install required libraries for their iotools usage.
"""


def _error_msg_one(*args):
    return "The module '{}' requires the library '{}' which is not installed.".format(*args)


def _error_msg_sev(*args):
    msg = "The module '{}' requires to install at least one of the following backends:" \
          "\n - pip install {}  # (recommended)" \
          "\n - pip install {}"
    args = list(args)
    for i, arg in enumerate(args):
        if not isinstance(arg, str):
            args[i] = " ".join(arg)
    return msg.format(args[0], args[1], "\n - pip install ".join(args[2:]))


def _warning_msg_default_one(*args):
    msg = "Warning: Function '{}' does not have the default behavior, " \
          "because library '{}' is not installed." \
          "\nPlease install it or use function '{}' instead."
    return msg.format(*args)


def _warning_msg_default_sev(func1, libs, func2):
    msg = "Warning: Function '{}' does not have the default behavior, " \
          "because libraries '{}' and '{}' are not installed." \
          "\nPlease install them or use function '{}' instead."
    return msg.format(func1, "', '".join(libs[:-1]), libs[-1], func2)


class _EmptyModule:

    def __init__(self, name, *other):
        self._name = name
        self._other = other

    def __call__(self, *args, **kwargs):
        if len(self._other) == 0:
            msg = "Module '{}' is required but not installed.\n(pip install {})"
            raise ModuleNotFoundError(msg.format(self._name, self._name))
        msg = "Modules '{}' and '{}' are required but not installed.\n(pip install {})"
        modules = [self._name] + list(self._other)
        raise ModuleNotFoundError(msg.format(", ".join(modules[:-1]), modules[-1], " ".join(modules)))

    def __getattr__(self, x):
        if len(self._other) == 0:
            msg = "Module '{}' is required but not installed.\n(pip install {})"
            raise ModuleNotFoundError(msg.format(self._name, self._name))
        msg = "Modules '{}' and '{}' are required but not installed.\n(pip install {})"
        modules = [self._name] + list(self._other)
        raise ModuleNotFoundError(msg.format(", ".join(modules[:-1]), modules[-1], " ".join(modules)))



"""
This module provides io for json files (.json extension)
See iotools.jsonio.help_json() for more info
"""

import json

from iotools._help_utils import _help_default
from iotools._generic_open import _generic_open


def decode_json(data, **kwargs):
    """
    Decode json data from str
    Backend used: json.loads (json package)

    Args:
        data (str): json data to decode
        **kwargs: same as in 'json.loads'

    Returns:
        dict | list: json data decoded
    """
    if isinstance(data, bytes):
        data = data.decode("utf-8")
    return json.loads(data, **kwargs)


def encode_json(data, **kwargs):
    """
    Encode json data to str
    Backend used: json.dumps (json package)

    Args:
        data (list | dict): json data to encode
        **kwargs: same as in 'json.dumps'

    Returns:
        str: json data encoded
    """
    return json.dumps(data, **kwargs)


def read_json(openfile, **kwargs):
    """
    Read a open json file
    Backend used: json.load (json package)

    Args:
        openfile (file-like): file to read, must have been opened
        **kwargs: same as in 'json.load'

    Returns:
        dict | list: data read from the file provided
    """
    return json.load(openfile, **kwargs)


def write_json(openfile, data, **kwargs):
    """
    Write in a open json file
    Backend used: json.dump (json package)

    Args:
        openfile (file-like): file to write, must have been opened
        data (list | dict): json data to save
        **kwargs: same as in 'json.dump'
    """
    json.dump(data, openfile, **kwargs)


def load_json(filename, fs=None, **kwargs):
    """
    Load a json file
    Backend used: json.load (json package)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        **kwargs: same as in 'json.load'

    Returns:
        dict | list: loaded json data
    """
    with _generic_open(filename, 'r', fs=fs) as f:
        data = read_json(f, **kwargs)
    return data


def save_json(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a json file
    Backend used: json.dump (json package)

    Args:
        filename (str): file name to save data to
        data (list | dict): json data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'json.dump'
    """
    with _generic_open(filename, 'w', fs=fs, makedirs=makedirs) as f:
        write_json(f, data, **kwargs)


def help_json():
    """
    Print help for json io
    """
    _help_default("json")




import os

from iotools.localfs import check_fs


def _generic_makedirs(filename, fs=None):
    fs = check_fs(fs)
    if isinstance(filename, str) and not fs.exists(os.path.dirname(filename)):
        fs.makedirs(os.path.dirname(filename))


def _generic_open(filename, *args, fs=None, makedirs=False, **kwargs):
    """
    Unified open, whether the filename provided is a str path, or a file-like object
    It does open, fs.open or nothing depending on filename and fs provided
    Args:
        filename (str | file_like): The file to open/return
        *args: will be passed on to "open" and "fs.open"
        fs: the file system to use (default is None)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: will be passed on to "open" and "fs.open"
    Returns:
        (file-like) : opened file
    """
    fs = check_fs(fs)
    if isinstance(filename, str):
        if makedirs and not fs.exists(os.path.dirname(filename)):
            fs.makedirs(os.path.dirname(filename))
        return fs.open(filename, *args, **kwargs)
    return filename  # file is already opened



"""
This module provides io for jpg files (.jpg extension)
See iotools.jpgio.help_jpg() for more info
"""

from iotools._help_utils import _help_default, _help_multi_packages
from iotools.imageio import (
    decode_image_using_imageio, encode_image_using_imageio,
    read_image_using_imageio, write_image_using_imageio,
    load_image_using_imageio, save_image_using_imageio,
    decode_image_using_skimage, encode_image_using_skimage,
    read_image_using_skimage, write_image_using_skimage,
    load_image_using_skimage, save_image_using_skimage,
    decode_image_using_pil, encode_image_using_pil,
    read_image_using_pil, write_image_using_pil,
    load_image_using_pil, save_image_using_pil,
    decode_image_using_cv2, encode_image_using_cv2,
    read_image_using_cv2, write_image_using_cv2,
    load_image_using_cv2, save_image_using_cv2,
    decode_image, encode_image, read_image, write_image, load_image, save_image
)

try:
    import imageio
except ModuleNotFoundError:
    from iotools._missing_module_helper import _EmptyModule
    imageio = _EmptyModule("imageio")


# +---------------+
# | Using imageio |
# +---------------+


def decode_jpg_using_imageio(data, **kwargs):
    return decode_image_using_imageio(data, **kwargs)


def encode_jpg_using_imageio(data, **kwargs):
    kwargs["format"] = "jpg"
    return encode_image_using_imageio(data, **kwargs)


def read_jpg_using_imageio(openfile, **kwargs):
    return read_image_using_imageio(openfile, **kwargs)


def write_jpg_using_imageio(openfile, data, **kwargs):
    kwargs["format"] = "jpg"
    return write_image_using_imageio(openfile, data, **kwargs)


def load_jpg_using_imageio(filename, fs=None, **kwargs):
    return load_image_using_imageio(filename, fs=fs, **kwargs)


def save_jpg_using_imageio(filename, data, fs=None, makedirs=False, **kwargs):
    kwargs["format"] = "jpg"
    return save_image_using_imageio(filename, data, fs=fs, makedirs=makedirs, **kwargs)


decode_jpg_using_imageio.__doc__ = decode_image_using_imageio.__doc__.replace(" image ", " jpg image ")
encode_jpg_using_imageio.__doc__ = encode_image_using_imageio.__doc__.replace(" image ", " jpg image ")
read_jpg_using_imageio.__doc__ = read_image_using_imageio.__doc__.replace(" image ", " jpg image ")
write_jpg_using_imageio.__doc__ = write_image_using_imageio.__doc__.replace(" image ", " jpg image ")
load_jpg_using_imageio.__doc__ = load_image_using_imageio.__doc__.replace(" image ", " jpg image ")
save_jpg_using_imageio.__doc__ = save_image_using_imageio.__doc__.replace(" image ", " jpg image ")


def help_jpg_using_imageio():
    """
    Print help for jpg image io using imageio library
    """
    _help_default("jpg_using_imageio")


# +---------------+
# | Using skimage |
# +---------------+


def decode_jpg_using_skimage(data, **kwargs):
    return decode_image_using_skimage(data, **kwargs)


def encode_jpg_using_skimage(data, **kwargs):
    kwargs["format"] = "jpg"
    return encode_image_using_skimage(data, **kwargs)


def read_jpg_using_skimage(openfile, **kwargs):
    return read_image_using_skimage(openfile, **kwargs)


def write_jpg_using_skimage(openfile, data, **kwargs):
    kwargs["format"] = "jpg"
    return write_image_using_skimage(openfile, data, **kwargs)


def load_jpg_using_skimage(filename, fs=None, **kwargs):
    return load_image_using_skimage(filename, fs=fs, **kwargs)


def save_jpg_using_skimage(filename, data, fs=None, makedirs=False, **kwargs):
    kwargs["format"] = "jpg"
    return save_image_using_skimage(filename, data, fs=fs, makedirs=makedirs, **kwargs)


decode_jpg_using_skimage.__doc__ = decode_image_using_skimage.__doc__.replace(" image ", " jpg image ")
encode_jpg_using_skimage.__doc__ = encode_image_using_skimage.__doc__.replace(" image ", " jpg image ")
read_jpg_using_skimage.__doc__ = read_image_using_skimage.__doc__.replace(" image ", " jpg image ")
write_jpg_using_skimage.__doc__ = write_image_using_skimage.__doc__.replace(" image ", " jpg image ")
load_jpg_using_skimage.__doc__ = load_image_using_skimage.__doc__.replace(" image ", " jpg image ")
save_jpg_using_skimage.__doc__ = save_image_using_skimage.__doc__.replace(" image ", " jpg image ")


def help_jpg_using_skimage():
    """
    Print help for jpg image io using skimage library
    """
    _help_default("jpg_using_skimage")


# +-----------+
# | Using PIL |
# +-----------+


def decode_jpg_using_pil(data, **kwargs):
    return decode_image_using_pil(data, **kwargs)


def encode_jpg_using_pil(data, **kwargs):
    kwargs["format"] = "jpg"
    return encode_image_using_pil(data, **kwargs)


def read_jpg_using_pil(openfile, **kwargs):
    return read_image_using_pil(openfile, **kwargs)


def write_jpg_using_pil(openfile, data, **kwargs):
    kwargs["format"] = "jpg"
    return write_image_using_pil(openfile, data, **kwargs)


def load_jpg_using_pil(filename, fs=None, **kwargs):
    return load_image_using_pil(filename, fs=fs, **kwargs)


def save_jpg_using_pil(filename, data, fs=None, makedirs=False, **kwargs):
    kwargs["format"] = "jpg"
    return save_image_using_pil(filename, data, fs=fs, makedirs=makedirs, **kwargs)


decode_jpg_using_pil.__doc__ = decode_image_using_pil.__doc__.replace(" image ", " jpg image ")
encode_jpg_using_pil.__doc__ = encode_image_using_pil.__doc__.replace(" image ", " jpg image ")
read_jpg_using_pil.__doc__ = read_image_using_pil.__doc__.replace(" image ", " jpg image ")
write_jpg_using_pil.__doc__ = write_image_using_pil.__doc__.replace(" image ", " jpg image ")
load_jpg_using_pil.__doc__ = load_image_using_pil.__doc__.replace(" image ", " jpg image ")
save_jpg_using_pil.__doc__ = save_image_using_pil.__doc__.replace(" image ", " jpg image ")


def help_jpg_using_pil():
    """
    Print help for jpg image io using pil library
    """
    _help_default("jpg_using_pil")


# +-----------+
# | Using cv2 |
# +-----------+


def decode_jpg_using_cv2(data, **kwargs):
    return decode_image_using_cv2(data, **kwargs)


def encode_jpg_using_cv2(data, **kwargs):
    kwargs["format"] = "jpg"
    return encode_image_using_cv2(data, **kwargs)


def read_jpg_using_cv2(openfile, **kwargs):
    return read_image_using_cv2(openfile, **kwargs)


def write_jpg_using_cv2(openfile, data, **kwargs):
    kwargs["format"] = "jpg"
    return write_image_using_cv2(openfile, data, **kwargs)


def load_jpg_using_cv2(filename, fs=None, **kwargs):
    return load_image_using_cv2(filename, fs=fs, **kwargs)


def save_jpg_using_cv2(filename, data, fs=None, makedirs=False, **kwargs):
    kwargs["format"] = "jpg"
    return save_image_using_cv2(filename, data, fs=fs, makedirs=makedirs, **kwargs)


decode_jpg_using_cv2.__doc__ = decode_image_using_cv2.__doc__.replace(" image ", " jpg image ")
encode_jpg_using_cv2.__doc__ = encode_image_using_cv2.__doc__.replace(" image ", " jpg image ")
read_jpg_using_cv2.__doc__ = read_image_using_cv2.__doc__.replace(" image ", " jpg image ")
write_jpg_using_cv2.__doc__ = write_image_using_cv2.__doc__.replace(" image ", " jpg image ")
load_jpg_using_cv2.__doc__ = load_image_using_cv2.__doc__.replace(" image ", " jpg image ")
save_jpg_using_cv2.__doc__ = save_image_using_cv2.__doc__.replace(" image ", " jpg image ")


def help_jpg_using_cv2():
    """
    Print help for jpg image io using cv2 library
    """
    _help_default("jpg_using_cv2")


# +------------------+
# | Default behavior |
# +------------------+


def decode_jpg(data, **kwargs):
    return decode_image(data, **kwargs)


def encode_jpg(data, **kwargs):
    kwargs["format"] = "jpg"
    return encode_image(data, **kwargs)


def read_jpg(openfile, **kwargs):
    return read_image(openfile, **kwargs)


def write_jpg(openfile, data, **kwargs):
    kwargs["format"] = "jpg"
    return write_image(openfile, data, **kwargs)


def load_jpg(filename, fs=None, **kwargs):
    return load_image(filename, fs=fs, **kwargs)


def save_jpg(filename, data, fs=None, makedirs=False, **kwargs):
    kwargs["format"] = "jpg"
    return save_image(filename, data, fs=fs, makedirs=makedirs, **kwargs)


decode_jpg.__doc__ = decode_image.__doc__.replace(" image ", " jpg image ")
encode_jpg.__doc__ = encode_image.__doc__.replace(" image ", " jpg image ")
read_jpg.__doc__ = read_image.__doc__.replace(" image ", " jpg image ")
write_jpg.__doc__ = write_image.__doc__.replace(" image ", " jpg image ")
load_jpg.__doc__ = load_image.__doc__.replace(" image ", " jpg image ")
save_jpg.__doc__ = save_image.__doc__.replace(" image ", " jpg image ")


def help_jpg():
    """
    Print help for jpg image io using several libraries
    """
    _help_multi_packages("jpg", "imageio, skimage, pil, cv2")



"""
This module provides io for zip files (.zip extension)
See iotools.zipio.help_zip() for more info
"""

import zipfile

from iotools._help_utils import _help_default
from iotools._generic_open import _generic_open
from iotools._default_decode_encode import _default_decode_bytes, _default_encode_bytes


def decode_zip(data, **kwargs):
    """
    Decode zip data from bytes
    Backend used: zipfile.ZipFile (zipfile package)

    Args:
        data (bytes): data to decode
        **kwargs: same as in 'zipfile.ZipFile'

    Returns:
        dict{filename: bytes}: zip data decoded
    """
    return _default_decode_bytes(read_zip, data, **kwargs)


def encode_zip(data, **kwargs):
    """
    Encode zip data to bytes
    Backend used: zipfile.ZipFile (zipfile package)

    Args:
        data (dict{filename: bytes}): zip data to encode
        **kwargs: same as in 'zipfile.ZipFile'

    Returns:
        bytes: zip data encoded
    """
    return _default_encode_bytes(write_zip, data, **kwargs)


def read_zip(openfile, **kwargs):
    """
    Read a open zip file
    Backend used: zipfile.ZipFile (zipfile package)

    Args:
        openfile (file-like): file to read, must have been opened
        **kwargs: same as in 'zipfile.ZipFile'

    Returns:
        dict{filename: bytes}: zip data read from the file provided
    """
    z = zipfile.ZipFile(openfile, **kwargs)
    data = {fn: None for fn in z.namelist() if not fn.endswith("/")}
    for subfilename in data:
        with z.open(subfilename, "r") as subf:
            data[subfilename] = subf.read()
    z.close()
    return data


def write_zip(openfile, data, **kwargs):
    """
    Write in a open zip file
    Backend used: zipfile.ZipFile (zipfile package)

    Args:
        openfile (file-like): file to write, must have been opened
        data (dict{filename: bytes}): zip data to save
        **kwargs: same as in 'zipfile.ZipFile'
    """
    z = zipfile.ZipFile(openfile, "w", **kwargs)
    for subfilename, raw_data in data.items():
        with z.open(subfilename, "w") as subf:
            if isinstance(raw_data, str):
                raw_data = raw_data.encode('utf-8')
            subf.write(raw_data)
    z.close()


def load_zip(filename, fs=None, **kwargs):
    """
    Load a zip file
    Backend used: zipfile.ZipFile (zipfile package)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        **kwargs: same as in 'zipfile.ZipFile'

    Returns:
        dict{filename: bytes}: zip data loaded from the file provided
    """
    with _generic_open(filename, 'rb', fs=fs) as f:
        data = read_zip(f, **kwargs)
    return data


def save_zip(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a zip file
    Backend used: zipfile.ZipFile (zipfile package)

    Args:
        filename (str): file name to save data to
        data (dict{filename: bytes}): zip data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'zipfile.ZipFile'
    """
    with _generic_open(filename, 'wb', fs=fs, makedirs=makedirs) as f:
        write_zip(f, data, **kwargs)


def help_zip():
    """
    Print help for zip io
    """
    _help_default("zip")



"""
This module provides io for gif videos (.gif extension)
See iotools.gifio.help_gif() for more info
"""

from iotools._help_utils import _help_default, _help_multi_packages
from iotools.videoio import (
    decode_video_using_imageio, encode_video_using_imageio,
    read_video_using_imageio, write_video_using_imageio,
    load_video_using_imageio, save_video_using_imageio,
    decode_video, encode_video, read_video, write_video, load_video, save_video
)


# +---------------+
# | Using imageio |
# +---------------+


def decode_gif_using_imageio(data, which="video", **kwargs):
    return decode_video_using_imageio(data, which=which, **kwargs)


def encode_gif_using_imageio(data, **kwargs):
    kwargs["format"] = "gif"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return encode_video_using_imageio(data, **kwargs)


def read_gif_using_imageio(openfile, which="video", **kwargs):
    return read_video_using_imageio(openfile, which=which, **kwargs)


def write_gif_using_imageio(openfile, data, **kwargs):
    kwargs["format"] = "gif"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return write_video_using_imageio(openfile, data, **kwargs)


def load_gif_using_imageio(filename, fs=None, which="video", **kwargs):
    return load_video_using_imageio(filename, fs=fs, which=which, **kwargs)


def save_gif_using_imageio(filename, data, fs=None, makedirs=False, **kwargs):
    kwargs["format"] = "gif"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return save_video_using_imageio(filename, data, fs=fs, makedirs=makedirs, **kwargs)


decode_gif_using_imageio.__doc__ = decode_video_using_imageio.__doc__.replace(" video ", " gif video ")
encode_gif_using_imageio.__doc__ = encode_video_using_imageio.__doc__.replace(" video ", " gif video ")
read_gif_using_imageio.__doc__ = read_video_using_imageio.__doc__.replace(" video ", " gif video ")
write_gif_using_imageio.__doc__ = write_video_using_imageio.__doc__.replace(" video ", " gif video ")
load_gif_using_imageio.__doc__ = load_video_using_imageio.__doc__.replace(" video ", " gif video ")
save_gif_using_imageio.__doc__ = save_video_using_imageio.__doc__.replace(" video ", " gif video ")


def help_gif_using_imageio():
    """
    Print help for gif video io using imageio library
    """
    _help_default("gif_using_imageio")


# +------------------+
# | Default behavior |
# +------------------+


def decode_gif(data, which="video", **kwargs):
    return decode_video(data, which=which, **kwargs)


def encode_gif(data, **kwargs):
    kwargs["format"] = "gif"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return encode_video(data, **kwargs)


def read_gif(openfile, which="video", **kwargs):
    return read_video(openfile, which=which, **kwargs)


def write_gif(openfile, data, **kwargs):
    kwargs["format"] = "gif"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return write_video(openfile, data, **kwargs)


def load_gif(filename, fs=None, which="video", **kwargs):
    return load_video(filename, fs=fs, which=which, **kwargs)


def save_gif(filename, data, fs=None, makedirs=False, **kwargs):
    kwargs["format"] = "gif"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return save_video(filename, data, fs=fs, makedirs=makedirs, **kwargs)


decode_gif.__doc__ = decode_video.__doc__.replace(" video ", " gif video ")
encode_gif.__doc__ = encode_video.__doc__.replace(" video ", " gif video ")
read_gif.__doc__ = read_video.__doc__.replace(" video ", " gif video ")
write_gif.__doc__ = write_video.__doc__.replace(" video ", " gif video ")
load_gif.__doc__ = load_video.__doc__.replace(" video ", " gif video ")
save_gif.__doc__ = save_video.__doc__.replace(" video ", " gif video ")


def help_gif():
    """
    Print help for gif video io using several libraries
    """
    _help_multi_packages("gif", "imageio, cv2")




import os
import random
import string
import tempfile

TEMP_FILENAME = os.path.join(tempfile.gettempdir(), "iotools_")


def _get_temp_filename(format=None):
    """
    Returns a temporary filename
    """
    if isinstance(format, str):
        format = "." + format.lower()
    else:
        format = ""
    for _ in range(10):
        temp_filename = TEMP_FILENAME + "".join(random.choice(string.ascii_letters) for i in range(8)) + format
        if not os.path.exists(temp_filename):
            return temp_filename
    raise OSError("Could not find a non-used temporary file")



"""
This module provides io for avi videos (.avi extension)
See iotools.aviio.help_avi() for more info
"""

from iotools._help_utils import _help_default, _help_multi_packages
from iotools.videoio import (
    decode_video_using_imageio, encode_video_using_imageio,
    read_video_using_imageio, write_video_using_imageio,
    load_video_using_imageio, save_video_using_imageio,
    decode_video_using_cv2, encode_video_using_cv2,
    read_video_using_cv2, write_video_using_cv2,
    load_video_using_cv2, save_video_using_cv2,
    decode_video, encode_video, read_video, write_video, load_video, save_video
)


# +---------------+
# | Using imageio |
# +---------------+


def decode_avi_using_imageio(data, which="video", **kwargs):
    return decode_video_using_imageio(data, which=which, **kwargs)


def encode_avi_using_imageio(data, **kwargs):
    kwargs["format"] = "avi"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return encode_video_using_imageio(data, **kwargs)


def read_avi_using_imageio(openfile, which="video", **kwargs):
    return read_video_using_imageio(openfile, which=which, **kwargs)


def write_avi_using_imageio(openfile, data, **kwargs):
    kwargs["format"] = "avi"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return write_video_using_imageio(openfile, data, **kwargs)


def load_avi_using_imageio(filename, fs=None, which="video", **kwargs):
    return load_video_using_imageio(filename, fs=fs, which=which, **kwargs)


def save_avi_using_imageio(filename, data, fs=None, makedirs=False, **kwargs):
    kwargs["format"] = "avi"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return save_video_using_imageio(filename, data, fs=fs, makedirs=makedirs, **kwargs)


decode_avi_using_imageio.__doc__ = decode_video_using_imageio.__doc__.replace(" video ", " avi video ")
encode_avi_using_imageio.__doc__ = encode_video_using_imageio.__doc__.replace(" video ", " avi video ")
read_avi_using_imageio.__doc__ = read_video_using_imageio.__doc__.replace(" video ", " avi video ")
write_avi_using_imageio.__doc__ = write_video_using_imageio.__doc__.replace(" video ", " avi video ")
load_avi_using_imageio.__doc__ = load_video_using_imageio.__doc__.replace(" video ", " avi video ")
save_avi_using_imageio.__doc__ = save_video_using_imageio.__doc__.replace(" video ", " avi video ")


def help_avi_using_imageio():
    """
    Print help for avi video io using imageio library
    """
    _help_default("avi_using_imageio")


# +-----------+
# | Using cv2 |
# +-----------+


def decode_avi_using_cv2(data, which="video", **kwargs):
    return decode_video_using_cv2(data, which=which, **kwargs)


def encode_avi_using_cv2(data, **kwargs):
    kwargs["format"] = "avi"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return encode_video_using_cv2(data, **kwargs)


def read_avi_using_cv2(openfile, which="video", **kwargs):
    return read_video_using_cv2(openfile, which=which, **kwargs)


def write_avi_using_cv2(openfile, data, **kwargs):
    kwargs["format"] = "avi"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return write_video_using_cv2(openfile, data, **kwargs)


def load_avi_using_cv2(filename, fs=None, which="video", **kwargs):
    return load_video_using_cv2(filename, fs=fs, which=which, **kwargs)


def save_avi_using_cv2(filename, data, fs=None, makedirs=False, **kwargs):
    kwargs["format"] = "avi"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return save_video_using_cv2(filename, data, fs=fs, makedirs=makedirs, **kwargs)


decode_avi_using_cv2.__doc__ = decode_video_using_cv2.__doc__.replace(" video ", " avi video ")
encode_avi_using_cv2.__doc__ = encode_video_using_cv2.__doc__.replace(" video ", " avi video ")
read_avi_using_cv2.__doc__ = read_video_using_cv2.__doc__.replace(" video ", " avi video ")
write_avi_using_cv2.__doc__ = write_video_using_cv2.__doc__.replace(" video ", " avi video ")
load_avi_using_cv2.__doc__ = load_video_using_cv2.__doc__.replace(" video ", " avi video ")
save_avi_using_cv2.__doc__ = save_video_using_cv2.__doc__.replace(" video ", " avi video ")


def help_avi_using_cv2():
    """
    Print help for avi video io using cv2 library
    """
    _help_default("avi_using_cv2")


# +------------------+
# | Default behavior |
# +------------------+


def decode_avi(data, which="video", **kwargs):
    return decode_video(data, which=which, **kwargs)


def encode_avi(data, **kwargs):
    kwargs["format"] = "avi"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return encode_video(data, **kwargs)


def read_avi(openfile, which="video", **kwargs):
    return read_video(openfile, which=which, **kwargs)


def write_avi(openfile, data, **kwargs):
    kwargs["format"] = "avi"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return write_video(openfile, data, **kwargs)


def load_avi(filename, fs=None, which="video", **kwargs):
    return load_video(filename, fs=fs, which=which, **kwargs)


def save_avi(filename, data, fs=None, makedirs=False, **kwargs):
    kwargs["format"] = "avi"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return save_video(filename, data, fs=fs, makedirs=makedirs, **kwargs)


decode_avi.__doc__ = decode_video.__doc__.replace(" video ", " avi video ")
encode_avi.__doc__ = encode_video.__doc__.replace(" video ", " avi video ")
read_avi.__doc__ = read_video.__doc__.replace(" video ", " avi video ")
write_avi.__doc__ = write_video.__doc__.replace(" video ", " avi video ")
load_avi.__doc__ = load_video.__doc__.replace(" video ", " avi video ")
save_avi.__doc__ = save_video.__doc__.replace(" video ", " avi video ")


def help_avi():
    """
    Print help for avi video io using several libraries
    """
    _help_multi_packages("avi", "imageio, cv2")



"""
This module provides io for videos (.avi, .gif and .mp4 extensions)
See iotools.videoio.help_video() for more info
"""

import os

from iotools.settings import settings
from iotools._default_decode_encode import _default_decode_bytes, _default_encode_bytes
from iotools._generic_open import _generic_open, _generic_makedirs
from iotools._help_utils import _help_default, _help_multi_packages
from iotools._missing_module_helper import _error_msg_sev, _EmptyModule, _warning_msg_default_sev
from iotools._temp_filename_utils import _get_temp_filename
from iotools.localfs import is_localfs

try:
    import numpy as np
except ModuleNotFoundError:
    from iotools._missing_module_helper import _error_msg_one
    if settings["SHOW_IMPORT_WARNINGS"]:
        print("Warning:", _error_msg_one("iotools.videoio", "numpy"))
    np = _EmptyModule("numpy")


_import_ok = set()

try:
    import imageio
    import imageio_ffmpeg  # noqa
    _import_ok.add("imageio")
except ModuleNotFoundError:
    imageio = _EmptyModule("imageio", "imageio-ffmpeg")

try:
    import cv2
    _import_ok.add("cv2")
except ModuleNotFoundError:
    cv2 = _EmptyModule("opencv-python")

if len(_import_ok) == 0:
    if settings["SHOW_IMPORT_WARNINGS"]:
        print("Warning:", _error_msg_sev("iotools.videoio", "imageio imageio-ffmpeg", "opencv-python"))


# +---------------+
# | Using imageio |
# +---------------+


DEFAULT_FORMAT = "mp4"
AVAILABLE_FORMATS = {"avi", "gif", "mp4"}
DEFAULT_PROPERTIES_IMAGEIO = {
    "avi": {"fps": 30},
    "gif": {"fps": 12},
    "mp4": {"fps": 30},
}
DEFAULT_PROPERTIES_CV2 = {
    "avi": {"fps": 30, "codec": "MJPG"},
    "gif": {"fps": 12, "codec": None},
    "mp4": {"fps": 30, "codec": "mp4v"},
}


class _Video(dict):
    pass


def _guess_format_from_filename(filename, default_format=DEFAULT_FORMAT):
    """
    Guess format from filename

    Args:
        filename (str | *): filename
        default_format (str): default format returned if format cannot be guessed from filename

    Returns:
        format (str): format guessed from filename. Returns 'default_format' if format cannot be guessed
    """
    if isinstance(filename, str) and ("." in filename):
        fmt = filename.rsplit(".", 1)[1].lower()
        if fmt in AVAILABLE_FORMATS:
            return fmt
    return default_format


def _insert_properties_in_data(filename, data, kwargs, default_properties):
    """
    Insert properties from kwargs to data
    """
    res = _Video()
    prop = {}
    prop["default_properties_used"] = set()
    properties_used = set()
    if isinstance(data, dict):
        data = data.copy()
    else:
        data = {"video": data}
    for key in {"video", "audio"}:
        res[key] = data.get(key, None)
    if "properties" not in data:
        data["properties"] = {}
    # handle format
    # get format from data or kwargs
    if "format" in kwargs:
        if "format" in data["properties"]:
            print("Warning: format provided twice.")
            properties_used.add("format")
        prop["format"] = kwargs.pop("format")
    elif "format" in data["properties"]:
        prop["format"] = data["properties"]["format"]
        properties_used.add("format")
    else:
        prop["format"] = _guess_format_from_filename(filename)
    prop["format"] = prop["format"].strip('.').lower()
    # check if format is available
    if prop["format"] not in AVAILABLE_FORMATS:
        print(
            "Warning: format '{}' is unknown or unavailable yet.".format(prop["format"]),
            "Using format '{}' instead".format(DEFAULT_FORMAT)
        )
        prop["format"] = DEFAULT_FORMAT
    # handle properties for each format
    for prop_name, prop_default in default_properties.get(prop["format"], {}).items():
        if prop_name in kwargs:
            if prop_name in data["properties"]:
                print("Warning: property '{}' provided twice.".format(prop_name))
                properties_used.add(prop_name)
            prop[prop_name] = kwargs.pop(prop_name)
        elif prop_name in data["properties"]:
            prop[prop_name] = data["properties"][prop_name]
            properties_used.add(prop_name)
        else:
            prop[prop_name] = prop_default
            prop["default_properties_used"].add(prop_name)
    # for prop_name in data["properties"].keys():
    #     if prop_name not in properties_used:
    #         print("Warning: ignoring unknown property '{}'.".format(prop_name))
    res["properties"] = prop
    return res, kwargs


###


def _imageio_properties_to_format(properties):
    if properties.get("version", b"").startswith(b"GIF"):
        return "gif"
    if properties.get("codec", "").startswith("gif"):
        return "gif"
    if properties.get("codec", "").startswith("mjpeg"):
        return "avi"
    if properties.get("codec", "").startswith("mpeg4"):
        return "mp4"
    if properties.get("codec", "").startswith("h26"):
        return "mp4"
    return "unknown"


def _encode_video_using_imageio(data, kwargs):
    prop = data["properties"]
    if settings["SHOW_DISK_USE_WARNINGS"]:
        print(
            "Warning: writing a '{}' video using imageio creates a temporary file".format(prop["format"])
            + " on local hard drive (it is removed automatically)."
        )
    temp_filename = _get_temp_filename(prop["format"])
    _save_video_using_imageio(temp_filename, data, kwargs)
    with open(temp_filename, "rb") as f:
        data_encoded = f.read()
    os.remove(temp_filename)
    return data_encoded


def _save_video_using_imageio(filename, data, kwargs):
    prop = kwargs.copy()
    prop.update(data["properties"])
    del prop["default_properties_used"]
    if prop["format"] == "gif" and "fps" in prop:
        prop["duration"] = 1000 * 1 / prop.pop("fps")
    writer = imageio.get_writer(filename, **prop)
    for frame in data["video"]:
        writer.append_data(frame)
    writer.close()


def decode_video_using_imageio(data, which="video", **kwargs):
    """
    Decode video data from bytes
    Backend used: imageio.get_reader (imageio package)

    Args:
        data (bytes): video data to decode
        which (str | list): which data to load, either 'video', 'audio', 'properties' or 'all'.
            Can also be a list of these str (example: which=['audio', 'video']). Defaults to 'video'
        **kwargs: same as in 'imageio.get_reader'

    Returns:
        dict: video data decoded, as a dict containing the keys 'video', 'audio' and 'properties'.
            If 'which' is a single element (example: which='video'), the dict value is returned instead
    """
    return _default_decode_bytes(read_video_using_imageio, data, which=which, **kwargs)


def encode_video_using_imageio(data, **kwargs):
    """
    Encode video data to bytes
    Backend used: imageio.get_writer (imageio package)

    Args:
        data (dict | np.array): video data to encode
        **kwargs: same as in 'imageio.get_writer'

    Returns:
        bytes: video data encoded
    """
    if not isinstance(data, _Video):
        data, kwargs = _insert_properties_in_data(None, data, kwargs, DEFAULT_PROPERTIES_IMAGEIO)
    if data["properties"]["format"] == "gif":
        return _default_encode_bytes(write_video_using_imageio, data, **kwargs)
    return _encode_video_using_imageio(data, kwargs)


def read_video_using_imageio(openfile, which="video", **kwargs):
    """
    Read an open video file
    Backend used: imageio.get_reader (imageio package)

    Args:
        openfile (file-like): file to read, must have been opened
        which (str | list): which data to load, either 'video', 'audio', 'properties' or 'all'.
            Can also be a list of these str (example: which=['audio', 'video']). Defaults to 'video'
        **kwargs: same as in 'imageio.get_reader'

    Returns:
        dict: video data read, as a dict containing the keys 'video', 'audio' and 'properties'.
            If 'which' is a single element (example: which='video'), the dict value is returned instead
    """
    res = {}
    # Read file and get right format
    formats = ["mp4", "avi", "gif"]
    if "format" in kwargs:
        formats = [kwargs.pop("format")] + formats
    for format in formats:
        try:
            reader = imageio.get_reader(openfile, format=format, **kwargs)
            properties = reader.get_meta_data()
            properties["format"] = _imageio_properties_to_format(properties)
            if properties["format"] != format:
                reader = imageio.get_reader(openfile, format=properties["format"], **kwargs)
                properties = reader.get_meta_data()
                properties["format"] = _imageio_properties_to_format(properties)
            break
        except ValueError:
            pass
    # Format data for user
    if ("video" in which) or ("all" in which):
        res["video"] = [im for im in reader]  # list(reader) does not work
    if ("audio" in which) or ("all" in which):
        res["audio"] = None
    if ("properties" in which) or ("all" in which):
        res["properties"] = dict(properties)
    if which == "video":
        return res["video"]
    if which == "audio":
        return res["audio"]
    if which == "properties":
        return res["properties"]
    return res


def write_video_using_imageio(openfile, data, **kwargs):
    """
    Write in an open video file
    Backend used: imageio.get_writer (imageio package)

    Args:
        openfile (file-like): file to write, must have been opened
        data (dict | np.array): video data to write
        **kwargs: same as in 'imageio.get_writer'
    """
    if not isinstance(data, _Video):
        data, kwargs = _insert_properties_in_data(None, data, kwargs, DEFAULT_PROPERTIES_IMAGEIO)
    if data["properties"]["format"] == "gif":
        _save_video_using_imageio(openfile, data, kwargs)
    else:
        data_enc = _encode_video_using_imageio(data, kwargs)
        openfile.write(data_enc)


def load_video_using_imageio(filename, fs=None, which="video", **kwargs):
    """
    Load a video file
    Backend used: imageio.get_reader (imageio package)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        which (str | list): which data to load, either 'video', 'audio', 'properties' or 'all'.
            Can also be a list of these str (example: which=['audio', 'video']). Defaults to 'video'
        **kwargs: same as in 'imageio.get_reader'

    Returns:
        dict: video data loaded, as a dict containing the keys 'video', 'audio' and 'properties'.
            If 'which' is a single element (example: which='video'), the dict value is returned instead
    """
    if "format" not in kwargs:
        fmt = _guess_format_from_filename(filename, default_format=None)
        if fmt is not None:
            kwargs["format"] = fmt
    # Better to guess the format from the filename to give info to imageio
    with _generic_open(filename, 'rb', fs=fs) as f:
        data = read_video_using_imageio(f, which=which, **kwargs)
    return data


def save_video_using_imageio(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a video file
    Backend used: imageio.get_writer (imageio package)

    Args:
        filename (str): file name to save data to
        data (dict | np.array): video data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'imageio.get_writer'
    """
    data, kwargs = _insert_properties_in_data(filename, data, kwargs, DEFAULT_PROPERTIES_IMAGEIO)
    if is_localfs(fs):
        if makedirs:
            _generic_makedirs(filename)
        _save_video_using_imageio(filename, data, kwargs)
    else:
        with _generic_open(filename, 'wb', fs=fs, makedirs=makedirs) as f:
            write_video_using_imageio(f, data, **kwargs)


def help_video_using_imageio():
    """
    Print help for video io using imageio library
    """
    _help_default("video_using_imageio")


# +-----------+
# | Using cv2 |
# +-----------+


def _get_video_cv2_properties(cap):
    properties = {"fps": cap.get(cv2.CAP_PROP_FPS)}
    h = int(cap.get(cv2.CAP_PROP_FOURCC))
    codec = chr(h & 0xff) + chr((h >> 8) & 0xff) + chr((h >> 16) & 0xff) + chr((h >> 24) & 0xff)
    properties["codec"] = codec.lower()
    return properties


def _cv2_properties_to_format(properties):
    if properties.get("codec", "").startswith("gif"):
        return "gif"
    if properties.get("codec", "") == "mjpg":
        return "avi"
    if properties.get("codec", "") in {"fmp4", "h264", "avc1", "mp4v"}:
        return "mp4"
    return "unknown"


def decode_video_using_cv2(data, which="video", **kwargs):
    """
    Decode video data from bytes
    Backend used: cv2.VideoCapture (opencv-python package)

    Args:
        data (bytes): video data to decode
        which (str | list): which data to load, either 'video', 'audio', 'properties' or 'all'.
            Can also be a list of these str (example: which=['audio', 'video']). Defaults to 'video'
        **kwargs: same as in 'cv2.VideoCapture'

    Returns:
        dict: video data decoded, as a dict containing the keys 'video', 'audio' and 'properties'.
            If 'which' is a single element (example: which='video'), the dict value is returned instead
    """
    if settings["SHOW_DISK_USE_WARNINGS"]:
        print(
            "Warning: reading a video using cv2 creates a temporary file on local hard drive"
            + " (it is removed automatically)."
        )
    temp_filename = _get_temp_filename()
    with open(temp_filename, "wb") as f:
        f.write(data)
    prop = {k: v for k, v in kwargs.items() if k not in {"codec", "format", "fps"}}
    data_decoded = load_video_using_cv2(temp_filename, fs=None, which=which, **prop)
    os.remove(temp_filename)
    return data_decoded


def encode_video_using_cv2(data, **kwargs):
    """
    Encode video data to bytes
    Backend used: cv2.VideoWriter (opencv-python package)

    Args:
        data (dict | np.array): video data to encode
        **kwargs: same as in 'cv2.VideoWriter'

    Returns:
        bytes: video data encoded
    """
    if not isinstance(data, _Video):
        data, kwargs = _insert_properties_in_data(None, data, kwargs, DEFAULT_PROPERTIES_CV2)
    if settings["SHOW_DISK_USE_WARNINGS"]:
        print(
            "Warning: writing a video using cv2 creates a temporary file on local hard drive"
            + " (it is removed automatically)."
        )
    temp_filename = _get_temp_filename(data["properties"]["format"])
    save_video_using_cv2(temp_filename, data, fs=None, makedirs=False, **kwargs)
    with open(temp_filename, "rb") as f:
        data_encoded = f.read()
    os.remove(temp_filename)
    return data_encoded


def read_video_using_cv2(openfile, which="video", **kwargs):
    """
    Read an open video file
    Backend used: cv2.VideoCapture (opencv-python package)

    Args:
        openfile (file-like): file to read, must have been opened
        which (str | list): which data to load, either 'video', 'audio', 'properties' or 'all'.
            Can also be a list of these str (example: which=['audio', 'video']). Defaults to 'video'
        **kwargs: same as in 'cv2.VideoCapture'

    Returns:
        dict: video data read, as a dict containing the keys 'video', 'audio' and 'properties'.
            If 'which' is a single element (example: which='video'), the dict value is returned instead
    """
    data_encoded = openfile.read()
    return decode_video_using_cv2(data_encoded, which=which, **kwargs)


def write_video_using_cv2(openfile, data, **kwargs):
    """
    Write in an open video file
    Backend used: cv2.VideoWriter (opencv-python package)

    Args:
        openfile (file-like): file to write, must have been opened
        data (dict | np.array): video data to write
        **kwargs: same as in 'cv2.VideoWriter'
    """
    if not isinstance(data, _Video):
        data, kwargs = _insert_properties_in_data(None, data, kwargs, DEFAULT_PROPERTIES_CV2)
    data_encoded = encode_video_using_cv2(data, **kwargs)
    openfile.write(data_encoded)


def load_video_using_cv2(filename, fs=None, which="video", **kwargs):
    """
    Load a video file
    Backend used: cv2.VideoCapture (opencv-python package)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        which (str | list): which data to load, either 'video', 'audio', 'properties' or 'all'.
            Can also be a list of these str (example: which=['audio', 'video']). Defaults to 'video'
        **kwargs: same as in 'cv2.VideoCapture'

    Returns:
        dict: video data loaded, as a dict containing the keys 'video', 'audio' and 'properties'.
            If 'which' is a single element (example: which='video'), the dict value is returned instead
    """
    if is_localfs(fs):
        # As far as I know a local filename is required to load a video using cv2
        if "format" in kwargs:
            del kwargs["format"]
        cap = cv2.VideoCapture(filename, **kwargs)
        res = {}
        if "properties" in which or "all" in which:
            res["properties"] = _get_video_cv2_properties(cap)
            if "format" not in res["properties"]:
                res["properties"]["format"] = _cv2_properties_to_format(res["properties"])
        if "video" in which or "all" in which:
            video = []
            ret = True
            while ret:
                ret, frame = cap.read()
                video.append(frame)
            cap.release()
            res["video"] = [frame[:, :, 2::-1] for frame in video[:-1]]
        if "audio" in which or "all" in which:
            res["audio"] = None
        if which == "video":
            return res["video"]
        if which == "audio":
            return res["audio"]
        if which == "properties":
            return res["properties"]
        return res
    # If file is not on local filesystem
    with _generic_open(filename, 'rb', fs=fs) as f:
        data = read_video_using_cv2(f, which=which, **kwargs)
    return data


def save_video_using_cv2(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a video file
    Backend used: cv2.VideoWriter (opencv-python package)

    Args:
        filename (str): file name to save data to
        data (dict | np.array): video data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'cv2.VideoWriter'
    """
    if not isinstance(data, _Video):
        data, kwargs = _insert_properties_in_data(filename, data, kwargs, DEFAULT_PROPERTIES_CV2)
    if is_localfs(fs):
        if makedirs:
            _generic_makedirs(filename)
        prop = data["properties"]
        if prop["codec"] is None:
            raise ValueError("Unknown format, please provide codec")
        frames = data["video"]
        height = len(frames[0])
        width = len(frames[0][0])
        fourcc = cv2.VideoWriter_fourcc(*prop["codec"])
        out = cv2.VideoWriter(filename, fourcc, prop["fps"], (width, height), **kwargs)
        for frame in frames:
            if frame.ndim != 3:
                raise ValueError("Frames must be 2D-images.")
            if frame.shape[2] != 3:
                raise ValueError("Frames must be 2D-images with 3 color channels.")
            if "uint8" not in str(frame.dtype):
                raise ValueError("Frames must be 2D-images with 3 color channels, of type uint8")
            out.write(frame[:, :, 2::-1])
        out.release()
    else:
        with _generic_open(filename, 'wb', fs=fs, makedirs=makedirs) as f:
            write_video_using_cv2(f, data, **kwargs)


def help_video_using_cv2():
    """
    Print help for video io using cv2 library
    """
    _help_default("video_using_cv2")


# +------------------+
# | Default behavior |
# +------------------+


def decode_video(data, which="video", **kwargs):
    """
    Decode video data from bytes
    It tries the following libraries (in the same order):
        imageio.get_reader
        cv2.VideoCapture
    If 'imageio.get_reader' does not work, a warning is printed since this function may have a
    different behavior in other environments

    Args:
        data (bytes): video data to decode
        which (str | list): which data to load, either 'video', 'audio', 'properties' or 'all'.
            Can also be a list of these str (example: which=['audio', 'video']). Defaults to 'video'
        **kwargs: same as in 'imageio.get_reader'

    Returns:
        dict: video data decoded, as a dict containing the keys 'video', 'audio' and 'properties'.
            If 'which' is a single element (example: which='video'), the dict value is returned instead
    """
    if not isinstance(imageio, _EmptyModule):
        return decode_video_using_imageio(data, which=which, **kwargs)
    if not isinstance(cv2, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_sev("decode_video", ["imageio", "imageio-ffmpeg"], "decode_video_using_cv2"))
        return decode_video_using_cv2(data, which=which, **kwargs)
    raise ModuleNotFoundError(_error_msg_sev("iotools.videoio", ["imageio", "imageio-ffmpeg"], "opencv-python"))


def encode_video(data, **kwargs):
    """
    Encode video data to bytes
    It tries the following libraries (in the same order):
        imageio.get_writer
        cv2.VideoWriter
    If 'imageio.get_writer' does not work, a warning is printed since this function may have a
    different behavior in other environments

    Args:
        data (dict | np.array): video data to encode
        **kwargs: same as in 'imageio.get_writer'

    Returns:
        bytes: video data encoded
    """
    if not isinstance(imageio, _EmptyModule):
        return encode_video_using_imageio(data, **kwargs)
    if not isinstance(cv2, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_sev("encode_video", ["imageio", "imageio-ffmpeg"], "encode_video_using_cv2"))
        return encode_video_using_cv2(data, **kwargs)
    raise ModuleNotFoundError(_error_msg_sev("iotools.videoio", ["imageio", "imageio-ffmpeg"], "opencv-python"))


def read_video(openfile, which="video", **kwargs):
    """
    Read an open video file
    It tries the following libraries (in the same order):
        imageio.get_reader
        cv2.VideoCapture
    If 'imageio.get_reader' does not work, a warning is printed since this function may have a
    different behavior in other environments

    Args:
        openfile (file-like): file to read, must have been opened
        which (str | list): which data to load, either 'video', 'audio', 'properties' or 'all'.
            Can also be a list of these str (example: which=['audio', 'video']). Defaults to 'video'
        **kwargs: same as in 'imageio.get_reader'

    Returns:
        dict: video data read, as a dict containing the keys 'video', 'audio' and 'properties'.
            If 'which' is a single element (example: which='video'), the dict value is returned instead
    """
    if not isinstance(imageio, _EmptyModule):
        return read_video_using_imageio(openfile, which=which, **kwargs)
    if not isinstance(cv2, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_sev("read_video", ["imageio", "imageio-ffmpeg"], "read_video_using_cv2"))
        return read_video_using_cv2(openfile, which=which, **kwargs)
    raise ModuleNotFoundError(_error_msg_sev("iotools.videoio", ["imageio", "imageio-ffmpeg"], "opencv-python"))


def write_video(openfile, data, **kwargs):
    """
    Write in an open video file
    It tries the following libraries (in the same order):
        imageio.get_writer
        cv2.VideoWriter
    If 'imageio.get_writer' does not work, a warning is printed since this function may have a
    different behavior in other environments

    Args:
        openfile (file-like): file to write, must have been opened
        data (dict | np.array): video data to write
        **kwargs: same as in 'imageio.get_writer'
    """
    if not isinstance(imageio, _EmptyModule):
        return write_video_using_imageio(openfile, data, **kwargs)
    if not isinstance(cv2, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_sev("write_video", ["imageio", "imageio-ffmpeg"], "write_video_using_cv2"))
        return write_video_using_cv2(openfile, data, **kwargs)
    raise ModuleNotFoundError(_error_msg_sev("iotools.videoio", ["imageio", "imageio-ffmpeg"], "opencv-python"))


def load_video(filename, fs=None, which="video", **kwargs):
    """
    Load a video file
    It tries the following libraries (in the same order):
        imageio.get_reader
        cv2.VideoCapture
    If 'imageio.get_reader' does not work, a warning is printed since this function may have a
    different behavior in other environments

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        which (str | list): which data to load, either 'video', 'audio', 'properties' or 'all'.
            Can also be a list of these str (example: which=['audio', 'video']). Defaults to 'video'
        **kwargs: same as in 'imageio.get_reader'

    Returns:
        dict: video data loaded, as a dict containing the keys 'video', 'audio' and 'properties'.
            If 'which' is a single element (example: which='video'), the dict value is returned instead
    """
    if not isinstance(imageio, _EmptyModule):
        return load_video_using_imageio(filename, fs=fs, which=which, **kwargs)
    if not isinstance(cv2, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_sev("load_video", ["imageio", "imageio-ffmpeg"], "load_video_using_cv2"))
        return load_video_using_cv2(filename, fs=fs, which=which, **kwargs)
    raise ModuleNotFoundError(_error_msg_sev("iotools.videoio", ["imageio", "imageio-ffmpeg"], "opencv-python"))


def save_video(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a video file
    It tries the following libraries (in the same order):
        imageio.get_writer
        cv2.VideoWriter
    If 'imageio.get_writer' does not work, a warning is printed since this function may have a
    different behavior in other environments

    Args:
        filename (str): file name to save data to
        data (dict | np.array): video data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'imageio.get_writer'
    """
    if not isinstance(imageio, _EmptyModule):
        return save_video_using_imageio(filename, data, fs=fs, makedirs=makedirs, **kwargs)
    if not isinstance(cv2, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_sev("save_video", ["imageio", "imageio-ffmpeg"], "save_video_using_cv2"))
        return save_video_using_cv2(filename, data, fs=fs, makedirs=makedirs, **kwargs)
    raise ModuleNotFoundError(_error_msg_sev("iotools.videoio", ["imageio", "imageio-ffmpeg"], "opencv-python"))


def help_video():
    """
    Print help for video io using several libraries
    """
    _help_multi_packages("video", "imageio, cv2")



"""
This module provides io for text files (.txt extension)
See iotools.txtio.help_txt() for more info
"""

from iotools._help_utils import _help_default
from iotools._generic_open import _generic_open


def decode_txt(data, **kwargs):
    """
    Decode text data
    Backend used: str.decode (builtin)

    Args:
        data (str | bytes): text data to decode
        **kwargs: same as in 'bytes.decode'

    Returns:
        str: decoded text data
    """
    if isinstance(data, bytes):
        return data.decode(**kwargs)
    return data


def encode_txt(data, **kwargs):
    """
    Encode text data
    Backend used: str.encode (builtin)

    Args:
        data (str | bytes): text data to encode
        **kwargs: same as in 'str.encode'

    Returns:
        str: encoded text data
    """
    if isinstance(data, bytes):
        return data
    return data.encode(**kwargs)


def read_txt(openfile, **kwargs):
    """
    Read an open text file
    Backend used: TextIOWrapper.read (builtin)

    Args:
        openfile (file-like): file to read, must have been opened
        **kwargs: same as in 'TextIOWrapper.read'

    Returns:
        str: text data read from the file provided
    """
    return openfile.read(**kwargs)


def write_txt(openfile, data, **kwargs):
    """
    Write in an open text file
    Backend used: TextIOWrapper.write (builtin)

    Args:
        openfile (file-like): file to write, must have been opened
        data (str): text data to save
        **kwargs: same as in 'TextIOWrapper.write'
    """
    openfile.write(data, **kwargs)


def load_txt(filename, fs=None, **kwargs):
    """
    Load a text file
    Backend used: TextIOWrapper.read (builtin)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        **kwargs: same as in 'TextIOWrapper.read'

    Returns:
        str: loaded text data
    """
    with _generic_open(filename, "r", fs=fs) as f:
        data = read_txt(f, **kwargs)
    return data


def save_txt(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a text file
    Backend used: TextIOWrapper.write (builtin)

    Args:
        filename (str): file name to save data to
        data (str): text data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'TextIOWrapper.write'
    """
    with _generic_open(filename, "w", fs=fs, makedirs=makedirs) as f:
        write_txt(f, data, **kwargs)


def help_txt():
    """
    Print help for txt io
    """
    _help_default("txt")



"""
This module provides io for zstandard files (.zst, .zstd extensions)
See iotools.zstio.help_zst() for more info
"""

from iotools._help_utils import _help_default
from iotools._generic_open import _generic_open
from iotools._default_decode_encode import _default_decode_bytes, _default_encode_bytes

try:
    import zstandard
except ModuleNotFoundError:
    from iotools.settings import settings
    from iotools._missing_module_helper import _EmptyModule, _error_msg_one
    if settings["SHOW_IMPORT_WARNINGS"]:
        print("Warning:", _error_msg_one("iotools.zstio", "zstandard"))
    zst = _EmptyModule("zstandard")


def decode_zst(data, **kwargs):
    """
    Decode zst data from str
    Backend used: zstandard.ZstdDecompressor (zstandard package)

    Args:
        data (bytes): data to decode
        **kwargs: same as in 'zstandard.ZstdDecompressor'

    Returns:
        bytes: zst data decoded
    """
    return _default_decode_bytes(read_zst, data, **kwargs)


def encode_zst(data, **kwargs):
    """
    Encode zst data to str
    Backend used: zstandard.ZstdCompressor (zstandard package)

    Args:
        data (bytes | str): zst data to encode
        **kwargs: same as in 'zstandard.ZstdCompressor'

    Returns:
        bytes: zst data encoded
    """
    return _default_encode_bytes(write_zst, data, **kwargs)


def read_zst(openfile, **kwargs):
    """
    Read a open zst file
    Backend used: zstandard.ZstdDecompressor (zstandard package)

    Args:
        openfile (file-like): file to read, must have been opened
        **kwargs: same as in 'zstandard.ZstdDecompressor'

    Returns:
        bytes: zst data read from the file provided
    """
    dctx = zstandard.ZstdDecompressor(**kwargs)
    reader = dctx.stream_reader(openfile)
    data = reader.read()
    return data


def write_zst(openfile, data, **kwargs):
    """
    Write in a open zst file
    Backend used: zstandard.ZstdCompressor (zstandard package)

    Args:
        openfile (file-like): file to write, must have been opened
        data (bytes | str): zst data to save
        **kwargs: same as in 'zstandard.ZstdCompressor'
    """
    ctx = zstandard.ZstdCompressor(**kwargs)
    writer = ctx.stream_writer(openfile)
    if isinstance(data, str):
        data = data.encode('utf-8')
    writer.write(data)
    writer.flush(zstandard.FLUSH_FRAME)
    openfile.flush()


def load_zst(filename, fs=None, **kwargs):
    """
    Load a zst file
    Backend used: zstandard.ZstdDecompressor (zstandard package)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        **kwargs: same as in 'zstandard.ZstdDecompressor'

    Returns:
        bytes: loaded zst data
    """
    with _generic_open(filename, 'rb', fs=fs) as f:
        data = read_zst(f, **kwargs)
    return data


def save_zst(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a zst file
    Backend used: zstandard.ZstdCompressor (zstandard package)

    Args:
        filename (str): file name to save data to
        data (bytes | str): zst data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'zstandard.ZstdCompressor'
    """
    with _generic_open(filename, 'wb', fs=fs, makedirs=makedirs) as f:
        write_zst(f, data, **kwargs)


def help_zst():
    """
    Print help for zst io
    """
    _help_default("zst")




import os

ROOT_PATH = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
EXAMPLES_PATH = os.path.join(ROOT_PATH, "examples")
SIMPLIFIED_IOTOOLS_PATH = os.path.join(ROOT_PATH, "simplified_iotools")


def _help_default(extension):
    examples_path = os.path.join(EXAMPLES_PATH, "examples_{}.py".format(extension))
    if "_" in extension:
        ext, using = extension.split("_", 1)
        simplified_iotools_path = os.path.join(SIMPLIFIED_IOTOOLS_PATH, ext + "io_" + using + ".py")
    else:
        simplified_iotools_path = os.path.join(SIMPLIFIED_IOTOOLS_PATH, extension + "io.py")
    with open(examples_path, "r") as f:
        example_code = f.read()
    with open(simplified_iotools_path, "r") as f:
        simplified_code = f.read()
    print("\n*** iotools simplified code ***\n\n{}\n".format(simplified_code))
    print("\n*** Examples ***\n\n{}\n".format(example_code))


def _help_multi_packages(extension, other_packages):
    _help_default(extension)
    print(
        "*** Misc ***\n\nOther packages for image io may be available,",
        "check help_{}_using_[{}]() for more info.".format(extension, other_packages)
    )




from io import StringIO, BytesIO


def _default_decode_str(func, data, **kwargs):
    filelike = StringIO(data)
    ret = func(filelike, **kwargs)
    filelike.close()
    return ret


def _default_encode_str(func, data, **kwargs):
    filelike = StringIO()
    func(filelike, data, **kwargs)
    filelike.seek(0)
    ret = filelike.read()
    filelike.close()
    return ret


def _default_decode_bytes(func, data, **kwargs):
    filelike = BytesIO(data)
    ret = func(filelike, **kwargs)
    filelike.close()
    return ret


def _default_encode_bytes(func, data, **kwargs):
    filelike = BytesIO()
    func(filelike, data, **kwargs)
    filelike.seek(0)
    ret = filelike.read()
    filelike.close()
    return ret


def _default_encode_bytes_or_str(func, data, **kwargs):
    try:
        ret = _default_encode_bytes(func, data, **kwargs)
    except TypeError as e:
        if e.args[0] == "a bytes-like object is required, not 'str'":
            ret = _default_encode_str(func, data, **kwargs)
    return ret



"""
This module provides io for mp4 videos (.mp4 extension)
See iotools.mp4io.help_mp4() for more info
"""

from iotools._help_utils import _help_default, _help_multi_packages
from iotools.videoio import (
    decode_video_using_imageio, encode_video_using_imageio,
    read_video_using_imageio, write_video_using_imageio,
    load_video_using_imageio, save_video_using_imageio,
    decode_video_using_cv2, encode_video_using_cv2,
    read_video_using_cv2, write_video_using_cv2,
    load_video_using_cv2, save_video_using_cv2,
    decode_video, encode_video, read_video, write_video, load_video, save_video
)


# +---------------+
# | Using imageio |
# +---------------+


def decode_mp4_using_imageio(data, which="video", **kwargs):
    return decode_video_using_imageio(data, which=which, **kwargs)


def encode_mp4_using_imageio(data, **kwargs):
    kwargs["format"] = "mp4"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return encode_video_using_imageio(data, **kwargs)


def read_mp4_using_imageio(openfile, which="video", **kwargs):
    return read_video_using_imageio(openfile, which=which, **kwargs)


def write_mp4_using_imageio(openfile, data, **kwargs):
    kwargs["format"] = "mp4"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return write_video_using_imageio(openfile, data, **kwargs)


def load_mp4_using_imageio(filename, fs=None, which="video", **kwargs):
    return load_video_using_imageio(filename, fs=fs, which=which, **kwargs)


def save_mp4_using_imageio(filename, data, fs=None, makedirs=False, **kwargs):
    kwargs["format"] = "mp4"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return save_video_using_imageio(filename, data, fs=fs, makedirs=makedirs, **kwargs)


decode_mp4_using_imageio.__doc__ = decode_video_using_imageio.__doc__.replace(" video ", " mp4 video ")
encode_mp4_using_imageio.__doc__ = encode_video_using_imageio.__doc__.replace(" video ", " mp4 video ")
read_mp4_using_imageio.__doc__ = read_video_using_imageio.__doc__.replace(" video ", " mp4 video ")
write_mp4_using_imageio.__doc__ = write_video_using_imageio.__doc__.replace(" video ", " mp4 video ")
load_mp4_using_imageio.__doc__ = load_video_using_imageio.__doc__.replace(" video ", " mp4 video ")
save_mp4_using_imageio.__doc__ = save_video_using_imageio.__doc__.replace(" video ", " mp4 video ")


def help_mp4_using_imageio():
    """
    Print help for mp4 video io using imageio library
    """
    _help_default("mp4_using_imageio")


# +-----------+
# | Using cv2 |
# +-----------+


def decode_mp4_using_cv2(data, which="video", **kwargs):
    return decode_video_using_cv2(data, which=which, **kwargs)


def encode_mp4_using_cv2(data, **kwargs):
    kwargs["format"] = "mp4"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return encode_video_using_cv2(data, **kwargs)


def read_mp4_using_cv2(openfile, which="video", **kwargs):
    return read_video_using_cv2(openfile, which=which, **kwargs)


def write_mp4_using_cv2(openfile, data, **kwargs):
    kwargs["format"] = "mp4"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return write_video_using_cv2(openfile, data, **kwargs)


def load_mp4_using_cv2(filename, fs=None, which="video", **kwargs):
    return load_video_using_cv2(filename, fs=fs, which=which, **kwargs)


def save_mp4_using_cv2(filename, data, fs=None, makedirs=False, **kwargs):
    kwargs["format"] = "mp4"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return save_video_using_cv2(filename, data, fs=fs, makedirs=makedirs, **kwargs)


decode_mp4_using_cv2.__doc__ = decode_video_using_cv2.__doc__.replace(" video ", " mp4 video ")
encode_mp4_using_cv2.__doc__ = encode_video_using_cv2.__doc__.replace(" video ", " mp4 video ")
read_mp4_using_cv2.__doc__ = read_video_using_cv2.__doc__.replace(" video ", " mp4 video ")
write_mp4_using_cv2.__doc__ = write_video_using_cv2.__doc__.replace(" video ", " mp4 video ")
load_mp4_using_cv2.__doc__ = load_video_using_cv2.__doc__.replace(" video ", " mp4 video ")
save_mp4_using_cv2.__doc__ = save_video_using_cv2.__doc__.replace(" video ", " mp4 video ")


def help_mp4_using_cv2():
    """
    Print help for mp4 video io using cv2 library
    """
    _help_default("mp4_using_cv2")


# +------------------+
# | Default behavior |
# +------------------+


def decode_mp4(data, which="video", **kwargs):
    return decode_video(data, which=which, **kwargs)


def encode_mp4(data, **kwargs):
    kwargs["format"] = "mp4"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return encode_video(data, **kwargs)


def read_mp4(openfile, which="video", **kwargs):
    return read_video(openfile, which=which, **kwargs)


def write_mp4(openfile, data, **kwargs):
    kwargs["format"] = "mp4"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return write_video(openfile, data, **kwargs)


def load_mp4(filename, fs=None, which="video", **kwargs):
    return load_video(filename, fs=fs, which=which, **kwargs)


def save_mp4(filename, data, fs=None, makedirs=False, **kwargs):
    kwargs["format"] = "mp4"
    if isinstance(data, dict) and "properties" in data and "format" in data["properties"]:
        del data["properties"]["format"]
    return save_video(filename, data, fs=fs, makedirs=makedirs, **kwargs)


decode_mp4.__doc__ = decode_video.__doc__.replace(" video ", " mp4 video ")
encode_mp4.__doc__ = encode_video.__doc__.replace(" video ", " mp4 video ")
read_mp4.__doc__ = read_video.__doc__.replace(" video ", " mp4 video ")
write_mp4.__doc__ = write_video.__doc__.replace(" video ", " mp4 video ")
load_mp4.__doc__ = load_video.__doc__.replace(" video ", " mp4 video ")
save_mp4.__doc__ = save_video.__doc__.replace(" video ", " mp4 video ")


def help_mp4():
    """
    Print help for mp4 video io using several libraries
    """
    _help_multi_packages("mp4", "imageio, cv2")



"""
This module provides io for images (.jpg, .png extensions)
See iotools.imageio.help_image() for more info
"""

import os

from iotools.settings import settings
from iotools._default_decode_encode import _default_decode_bytes, _default_encode_bytes
from iotools._generic_open import _generic_open, _generic_makedirs
from iotools._help_utils import _help_default, _help_multi_packages
from iotools._missing_module_helper import _error_msg_sev, _EmptyModule, _warning_msg_default_one
from iotools._temp_filename_utils import _get_temp_filename
from iotools.localfs import is_localfs


DEFAULT_FORMAT = "png"
AVAILABLE_FORMATS = {"jpeg", "jpg", "png"}
DEFAULT_PROPERTIES = {
    "jpg": {"quality": 90},
}


class _Image(dict):
    pass


try:
    import numpy as np
except ModuleNotFoundError:
    from iotools._missing_module_helper import _error_msg_one
    if settings["SHOW_IMPORT_WARNINGS"]:
        print("Warning:", _error_msg_one("iotools.imageio", "numpy"))
    np = _EmptyModule("numpy")


_import_ok = set()

try:
    import imageio
    _import_ok.add("imageio")
except ModuleNotFoundError:
    imageio = _EmptyModule("imageio")

try:
    import skimage.io as skimageio
    _import_ok.add("skimage")
except ModuleNotFoundError:
    skimageio = _EmptyModule("scikit-image")

try:
    import PIL.Image as PilImage
    _import_ok.add("PilImage")
except ModuleNotFoundError:
    PilImage = _EmptyModule("Pillow")

try:
    import cv2
    _import_ok.add("cv2")
except ModuleNotFoundError:
    cv2 = _EmptyModule("opencv-python")


if len(_import_ok) == 0:
    if settings["SHOW_IMPORT_WARNINGS"]:
        print("Warning:", _error_msg_sev("iotools.imageio", "imageio", "scikit-image", "Pillow", "opencv-python"))


def _guess_format_from_filename(filename, default_format=DEFAULT_FORMAT):
    """
    Guess format from filename

    Args:
        filename (str | *): filename
        default_format (str): default format returned if format cannot be guessed from filename

    Returns:
        format (str): format guessed from filename. Returns 'default_format' if format cannot be guessed
    """
    if isinstance(filename, str) and ("." in filename):
        fmt = filename.rsplit(".", 1)[1].lower()
        if fmt in AVAILABLE_FORMATS:
            return fmt
    return default_format


def _insert_properties_in_data(filename, data, kwargs):
    """
    Insert properties from kwargs to data
    """
    res = _Image()
    res["default_properties_used"] = set()
    if isinstance(data, dict):
        res["image"] = data.pop("image")
    else:
        res["image"] = data
        data = {}
    # handle format
    if "format" in data:
        if "format" in kwargs:
            print("Warning: format provided twice.")
            kwargs.pop("format")
        res["format"] = data.pop("format")
    elif "format" in kwargs:
        res["format"] = kwargs.pop("format")
    else:
        res["format"] = _guess_format_from_filename(filename)
    res["format"] = res["format"].strip('.').lower()
    if res["format"] == "jpeg":
        res["format"] = "jpg"
    # handle properties for each format
    for prop_name, prop_default in DEFAULT_PROPERTIES.get(res["format"], {}).items():
        if prop_name in data:
            if prop_name in kwargs:
                print("Warning: property '{}' provided twice.".format(prop_name))
                kwargs.pop(prop_name)
            res[prop_name] = data.pop(prop_name)
        elif prop_name in kwargs:
            res[prop_name] = kwargs.pop(prop_name)
        else:
            res[prop_name] = prop_default
            res["default_properties_used"].add(prop_name)
    for prop in data.keys():
        print("Warning: ignoring unknown property '{}'.".format(prop))
    return res, kwargs


def _del_alpha_channel(img):
    """
    Deletes the alpha channel
    """
    if isinstance(img, np.ndarray) and (img.ndim == 3) and (img.shape[2] == 4):
        alpha = set(np.unique(img[:, :, 3]).tolist())
        if alpha != {255}:
            print("Warning: 'jpg' format does not support alpha channel. Ignoring alpha channel.")
        return img[:, :, :3]
    return img


# +---------------+
# | Using imageio |
# +---------------+


def decode_image_using_imageio(data, **kwargs):
    """
    Decode image data from bytes
    Backend used: imageio.v3.imread (imageio package)

    Args:
        data (bytes): image data to decode
        **kwargs: same as in 'imageio.v3.imread'

    Returns:
        array: image data decoded
    """
    return _default_decode_bytes(read_image_using_imageio, data, **kwargs)


def encode_image_using_imageio(data, **kwargs):
    """
    Encode image data to bytes
    Backend used: imageio.imsave (imageio package)

    Args:
        data (array): image data to encode
        **kwargs: same as in 'imageio.imsave'

    Returns:
        bytes: image data encoded
    """
    return _default_encode_bytes(write_image_using_imageio, data, **kwargs)


def read_image_using_imageio(openfile, **kwargs):
    """
    Read a open image file
    Backend used: imageio.v3.imread (imageio package)

    Args:
        openfile (file-like): file to read, must have been opened
        **kwargs: same as in 'imageio.v3.imread'

    Returns:
        array: image data read from the file provided
    """
    kwargs.pop("format", None)
    img = imageio.v3.imread(openfile, **kwargs)
    return img


def write_image_using_imageio(openfile, data, **kwargs):
    """
    Write in a open image file
    Backend used: imageio.imsave (imageio package)

    Args:
        openfile (file-like): file to write, must have been opened
        data (array): image data to save
        **kwargs: same as in 'imageio.imsave'
    """
    if not isinstance(data, _Image):
        data, kwargs = _insert_properties_in_data(openfile, data, kwargs)
    if data["format"] == "jpg":
        image = _del_alpha_channel(data["image"])
        imageio.imsave(openfile, image, format=data["format"], quality=data["quality"], **kwargs)
    else:
        imageio.imsave(openfile, data["image"], format=data["format"], **kwargs)


def load_image_using_imageio(filename, fs=None, **kwargs):
    """
    Load a image file
    Backend used: imageio.v3.imread (imageio package)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        **kwargs: same as in 'imageio.v3.imread'

    Returns:
        array: loaded data
    """
    with _generic_open(filename, 'rb', fs=fs) as f:
        data = read_image_using_imageio(f, **kwargs)
    return data


def save_image_using_imageio(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a image file
    Backend used: imageio.imsave (imageio package)

    Args:
        filename (str): file name to save data to
        data (array): image data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'imageio.imsave'
    """
    data, kwargs = _insert_properties_in_data(filename, data, kwargs)
    # When file is opened, image saver cannot infer the format from the file name
    with _generic_open(filename, 'wb', fs=fs, makedirs=makedirs) as f:
        write_image_using_imageio(f, data, **kwargs)


def help_image_using_imageio():
    """
    Print help for image io using imageio library
    """
    _help_default("image_using_imageio")


# +---------------+
# | Using skimage |
# +---------------+


def decode_image_using_skimage(data, **kwargs):
    """
    Decode image data from bytes
    Backend used: skimage.io.imread (scikit-image package)

    Args:
        data (bytes): image data to decode
        **kwargs: same as in 'skimage.io.imread'

    Returns:
        array: image data decoded
    """
    return _default_decode_bytes(read_image_using_skimage, data, **kwargs)


def encode_image_using_skimage(data, **kwargs):
    """
    Encode image data to bytes
    Backend used: skimage.io.imsave (scikit-image package)

    Args:
        data (array): image data to encode
        **kwargs: same as in 'skimage.io.imsave'

    Returns:
        bytes: image data encoded
    """
    return _default_encode_bytes(write_image_using_skimage, data, **kwargs)


def read_image_using_skimage(openfile, **kwargs):
    """
    Read a open image file
    Backend used: skimage.io.imread (scikit-image package)

    Args:
        openfile (file-like): file to read, must have been opened
        **kwargs: same as in 'skimage.io.imread'

    Returns:
        array: image data read from the file provided
    """
    kwargs.pop("format", None)
    img = skimageio.imread(openfile, **kwargs)
    return img


def write_image_using_skimage(openfile, data, **kwargs):
    """
    Write in a open image file
    Backend used: skimage.io.imsave (scikit-image package)

    Args:
        openfile (file-like): file to write, must have been opened
        data (array): image data to save
        **kwargs: same as in 'skimage.io.imsave'
    """
    if not isinstance(data, _Image):
        data, kwargs = _insert_properties_in_data(openfile, data, kwargs)
    if data["format"] == "jpg":
        image = _del_alpha_channel(data["image"])
        skimageio.imsave(openfile, image, extension="." + data["format"], quality=data["quality"], **kwargs)
    else:
        skimageio.imsave(openfile, data["image"], extension="." + data["format"], **kwargs)


def load_image_using_skimage(filename, fs=None, **kwargs):
    """
    Load a image file
    Backend used: skimage.io.imread (scikit-image package)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        **kwargs: same as in 'skimage.io.imread'

    Returns:
        array: loaded data
    """
    with _generic_open(filename, 'rb', fs=fs) as f:
        data = read_image_using_skimage(f, **kwargs)
    return data


def save_image_using_skimage(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a image file
    Backend used: skimage.io.imsave (scikit-image package)

    Args:
        filename (str): file name to save data to
        data (array): image data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'skimage.io.imsave'
    """
    data, kwargs = _insert_properties_in_data(filename, data, kwargs)
    # When file is opened, image saver cannot infer the format from the file name
    with _generic_open(filename, 'wb', fs=fs, makedirs=makedirs) as f:
        write_image_using_skimage(f, data, **kwargs)


def help_image_using_skimage():
    """
    Print help for image io using skimage library
    """
    _help_default("image_using_skimage")


# +-----------+
# | Using PIL |
# +-----------+


def decode_image_using_pil(data, **kwargs):
    """
    Decode image data from bytes
    Backend used: PIL.Image.open (Pillow package)

    Args:
        data (bytes): image data to decode
        **kwargs: same as in 'PIL.Image.open'

    Returns:
        array: image data decoded
    """
    return _default_decode_bytes(read_image_using_pil, data, **kwargs)


def encode_image_using_pil(data, **kwargs):
    """
    Encode image data to bytes
    Backend used: PIL.Image.save (Pillow package)

    Args:
        data (array): image data to encode
        **kwargs: same as in 'PIL.Image.save'

    Returns:
        bytes: image data encoded
    """
    return _default_encode_bytes(write_image_using_pil, data, **kwargs)


def read_image_using_pil(openfile, **kwargs):
    """
    Read a open image file
    Backend used: PIL.Image.open (Pillow package)

    Args:
        openfile (file-like): file to read, must have been opened
        **kwargs: same as in 'PIL.Image.open'

    Returns:
        array: image data read from the file provided
    """
    kwargs.pop("format", None)
    image = PilImage.open(openfile, **kwargs)
    image.load()
    return image


def write_image_using_pil(openfile, data, **kwargs):
    """
    Write in a open image file
    Backend used: PIL.Image.save (Pillow package)

    Args:
        openfile (file-like): file to write, must have been opened
        data (array): image data to save
        **kwargs: same as in 'PIL.Image.save'
    """
    if not isinstance(data, _Image):
        data, kwargs = _insert_properties_in_data(openfile, data, kwargs)
    if isinstance(data, np.ndarray):
        data["image"] = PilImage.fromarray(data["image"])
    if data["format"] == "jpg":
        data["format"] = "jpeg"
        data["image"].save(openfile, format=data["format"], quality=data["quality"], **kwargs)
    else:
        data["image"].save(openfile, format=data["format"], **kwargs)


def load_image_using_pil(filename, fs=None, **kwargs):
    """
    Load a image file
    Backend used: PIL.Image.open (Pillow package)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        **kwargs: same as in 'PIL.Image.open'

    Returns:
        array: loaded data
    """
    with _generic_open(filename, 'rb', fs=fs) as f:
        data = read_image_using_pil(f, **kwargs)
    return data


def save_image_using_pil(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a image file
    Backend used: PIL.Image.save (Pillow package)

    Args:
        filename (str): file name to save data to
        data (array): image data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'PIL.Image.save'
    """
    data, kwargs = _insert_properties_in_data(filename, data, kwargs)
    # When file is opened, image saver cannot infer the format from the file name
    with _generic_open(filename, 'wb', fs=fs, makedirs=makedirs) as f:
        write_image_using_pil(f, data, **kwargs)


def help_image_using_pil():
    """
    Print help for image io using PIL library
    """
    _help_default("image_using_pil")


# +-----------+
# | Using cv2 |
# +-----------+


def decode_image_using_cv2(data, **kwargs):
    """
    Decode image data from bytes
    Backend used: cv2.imdecode (opencv-python package)

    Args:
        data (bytes): image data to decode
        **kwargs: same as in 'cv2.imdecode'

    Returns:
        array: image data decoded
    """
    return cv2.imdecode(np.frombuffer(data, np.uint8), cv2.IMREAD_UNCHANGED, **kwargs)


def encode_image_using_cv2(data, **kwargs):
    """
    Encode image data to bytes
    Backend used: cv2.imencode (opencv-python package)

    Args:
        data (array): image data to encode
        **kwargs: same as in 'cv2.imencode'

    Returns:
        bytes: image data encoded
    """
    if not isinstance(data, _Image):
        data, kwargs = _insert_properties_in_data(None, data, kwargs)
    if data["format"] == "jpg":
        data["image"] = _del_alpha_channel(data["image"])
        res = cv2.imencode("." + data["format"], data["image"], [cv2.IMWRITE_JPEG_QUALITY, data["quality"]], **kwargs)
    else:
        res = cv2.imencode("." + data["format"], data["image"], **kwargs)
    return res[1].tobytes()


def read_image_using_cv2(openfile, **kwargs):
    """
    Read a open image file
    Backend used: cv2.imdecode (opencv-python package)

    Args:
        openfile (file-like): file to read, must have been opened
        **kwargs: same as in 'cv2.imdecode'

    Returns:
        array: image data read from the file provided
    """
    data_encoded = openfile.read()
    return decode_image_using_cv2(data_encoded, **kwargs)


def write_image_using_cv2(openfile, data, **kwargs):
    """
    Write in a open image file
    Backend used: cv2.imencode (opencv-python package)

    Args:
        openfile (file-like): file to write, must have been opened
        data (array): image data to save
        **kwargs: same as in 'cv2.imencode'
    """
    data_encoded = encode_image_using_cv2(data, **kwargs)
    openfile.write(data_encoded)


def load_image_using_cv2(filename, fs=None, **kwargs):
    """
    Load a image file
    Backend used: cv2.imread (opencv-python package)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        **kwargs: same as in 'cv2.imread'

    Returns:
        array: loaded data
    """
    if is_localfs(fs):
        kwargs.pop("format", None)
        data = cv2.imread(filename, **kwargs)
    else:
        with _generic_open(filename, 'rb', fs=fs) as f:
            data = read_image_using_cv2(f, **kwargs)
    return data


def save_image_using_cv2(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a image file
    Backend used: cv2.imwrite (opencv-python package)

    Args:
        filename (str): file name to save data to
        data (array): image data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'cv2.imwrite'
    """
    data, kwargs = _insert_properties_in_data(filename, data, kwargs)
    # When file is opened, image saver cannot infer the format from the file name
    if is_localfs(fs):
        if makedirs:
            _generic_makedirs(filename, fs=fs)
        if filename.endswith("." + data["format"]):
            temp_filename = filename
        else:
            temp_filename = _get_temp_filename(data["format"])
        if data["format"] == "jpg":
            data["image"] = _del_alpha_channel(data["image"])
            cv2.imwrite(temp_filename, data["image"], [cv2.IMWRITE_JPEG_QUALITY, data["quality"]], **kwargs)
        else:
            cv2.imwrite(temp_filename, data["image"], **kwargs)
        if temp_filename != filename:
            os.rename(temp_filename, filename)
    else:
        with _generic_open(filename, 'wb', fs=fs, makedirs=makedirs) as f:
            write_image_using_cv2(f, data, **kwargs)


def help_image_using_cv2():
    """
    Print help for image io using cv2 library
    """
    _help_default("image_using_cv2")


# +------------------+
# | Default behavior |
# +------------------+


def decode_image(data, **kwargs):
    """
    Decode image data from bytes
    It tries the following libraries (in the same order):
        imageio.imread
        skimage.io.imread
        PIL.Image.open
        cv2.imread
    If 'imageio.imread' does not work, a warning is printed since this function may have a
    different behavior in other environments

    Args:
        data (bytes): image data to decode
        **kwargs: same as in 'imageio.imread'

    Returns:
        array: image data decoded
    """
    if not isinstance(imageio, _EmptyModule):
        return decode_image_using_imageio(data, **kwargs)
    if not isinstance(skimageio, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_one("decode_image", "imageio", "decode_image_using_skimage"))
        return decode_image_using_skimage(data, **kwargs)
    if not isinstance(PilImage, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_one("decode_image", "imageio", "decode_image_using_pil"))
        return decode_image_using_pil(data, **kwargs)
    if not isinstance(cv2, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_one("decode_image", "imageio", "decode_image_using_cv2"))
        return decode_image_using_cv2(data, **kwargs)
    raise ModuleNotFoundError(
        _error_msg_sev("iotools.imageio", "imageio", "scikit-image", "Pillow", "opencv-python"))


def encode_image(data, **kwargs):
    """
    Encode image data to bytes
    It tries the following libraries (in the same order):
        imageio.imsave
        skimage.io.imsave
        PIL.Image.save
        cv2.imwrite
    If 'imageio.imsave' does not work, a warning is printed since this function may have a
    different behavior in other environments

    Args:
        data (array): image data to encode
        **kwargs: same as in 'imageio.imsave'

    Returns:
        bytes: image data encoded
    """
    if not isinstance(imageio, _EmptyModule):
        return encode_image_using_imageio(data, **kwargs)
    if not isinstance(skimageio, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_one("encode_image", "imageio", "encode_image_using_skimage"))
        return encode_image_using_skimage(data, **kwargs)
    if not isinstance(PilImage, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_one("encode_image", "imageio", "encode_image_using_pil"))
        return encode_image_using_pil(data, **kwargs)
    if not isinstance(cv2, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_one("encode_image", "imageio", "encode_image_using_cv2"))
        return encode_image_using_cv2(data, **kwargs)
    raise ModuleNotFoundError(
        _error_msg_sev("iotools.imageio", "imageio", "scikit-image", "Pillow", "opencv-python"))


def read_image(openfile, **kwargs):
    """
    Read a open image file
    It tries the following libraries (in the same order):
        imageio.imread
        skimage.io.imread
        PIL.Image.open
        cv2.imread
    If 'imageio.imread' does not work, a warning is printed since this function may have a
    different behavior in other environments

    Args:
        openfile (file-like): file to read, must have been opened
        **kwargs: same as in 'imageio.imread'

    Returns:
        array: image data read from the file provided
    """
    if not isinstance(imageio, _EmptyModule):
        return read_image_using_imageio(openfile, **kwargs)
    if not isinstance(skimageio, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_one("read_image", "imageio", "read_image_using_skimage"))
        return read_image_using_skimage(openfile, **kwargs)
    if not isinstance(PilImage, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_one("read_image", "imageio", "read_image_using_pil"))
        return read_image_using_pil(openfile, **kwargs)
    if not isinstance(cv2, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_one("read_image", "imageio", "read_image_using_cv2"))
        return read_image_using_cv2(openfile, **kwargs)
    raise ModuleNotFoundError(
        _error_msg_sev("iotools.imageio", "imageio", "scikit-image", "Pillow", "opencv-python"))


def write_image(openfile, data, **kwargs):
    """
    Write in a open image file
    It tries the following libraries (in the same order):
        imageio.imsave
        skimage.io.imsave
        PIL.Image.save
        cv2.imwrite
    If 'imageio.imsave' does not work, a warning is printed since this function may have a
    different behavior in other environments

    Args:
        openfile (file-like): file to write, must have been opened
        data (array): image data to save
        **kwargs: same as in 'imageio.imsave'
    """
    if not isinstance(imageio, _EmptyModule):
        return write_image_using_imageio(openfile, data, **kwargs)
    if not isinstance(skimageio, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_one("write_image", "imageio", "write_image_using_skimage"))
        return write_image_using_skimage(openfile, data, **kwargs)
    if not isinstance(PilImage, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_one("write_image", "imageio", "write_image_using_pil"))
        return write_image_using_pil(openfile, data, **kwargs)
    if not isinstance(cv2, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_one("write_image", "imageio", "write_image_using_cv2"))
        return write_image_using_cv2(openfile, data, **kwargs)
    raise ModuleNotFoundError(
        _error_msg_sev("iotools.imageio", "imageio", "scikit-image", "Pillow", "opencv-python"))


def load_image(filename, fs=None, **kwargs):
    """
    Load a image file
    It tries the following libraries (in the same order):
        imageio.imread
        skimage.io.imread
        PIL.Image.open
        cv2.imread
    If 'imageio.imread' does not work, a warning is printed since this function may have a
    different behavior in other environments

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        **kwargs: same as in 'imageio.imread'

    Returns:
        array: loaded data
    """
    if not isinstance(imageio, _EmptyModule):
        return load_image_using_imageio(filename, fs=fs, **kwargs)
    if not isinstance(skimageio, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_one("load_image", "imageio", "load_image_using_skimage"))
        return load_image_using_skimage(filename, fs=fs, **kwargs)
    if not isinstance(PilImage, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_one("load_image", "imageio", "load_image_using_pil"))
        return load_image_using_pil(filename, fs=fs, **kwargs)
    if not isinstance(cv2, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_one("load_image", "imageio", "load_image_using_cv2"))
        return load_image_using_cv2(filename, fs=fs, **kwargs)
    raise ModuleNotFoundError(
        _error_msg_sev("iotools.imageio", "imageio", "scikit-image", "Pillow", "opencv-python"))


def save_image(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a image file
    It tries the following libraries (in the same order):
        imageio.imsave
        skimage.io.imsave
        PIL.Image.save
        cv2.imwrite
    If 'imageio.imsave' does not work, a warning is printed since this function may have a
    different behavior in other environments

    Args:
        filename (str): file name to save data to
        data (array): image data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'imageio.imsave'
    """
    if not isinstance(imageio, _EmptyModule):
        return save_image_using_imageio(filename, data, fs=fs, makedirs=makedirs, **kwargs)
    if not isinstance(skimageio, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_one("save_image", "imageio", "save_image_using_skimage"))
        return save_image_using_skimage(filename, data, fs=fs, makedirs=makedirs, **kwargs)
    if not isinstance(PilImage, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_one("save_image", "imageio", "save_image_using_pil"))
        return save_image_using_pil(filename, data, fs=fs, makedirs=makedirs, **kwargs)
    if not isinstance(cv2, _EmptyModule):
        if settings["SHOW_DEFAULT_BEHAVIOR_WARNINGS"]:
            print(_warning_msg_default_one("save_image", "imageio", "save_image_using_cv2"))
        return save_image_using_cv2(filename, data, fs=fs, makedirs=makedirs, **kwargs)
    raise ModuleNotFoundError(
        _error_msg_sev("iotools.imageio", "imageio", "scikit-image", "Pillow", "opencv-python"))


def help_image():
    """
    Print help for image io using several libraries
    """
    _help_multi_packages("image", "imageio, skimage, pil, cv2")



"""
This module provides io for pickle files (.pickle extension)
See iotools.pickleio.help_pickle() for more info
"""

import pickle

from iotools._help_utils import _help_default
from iotools._generic_open import _generic_open


def decode_pickle(data, **kwargs):
    """
    Decode pickle data from bytes
    Backend used: pickle.loads (pickle package)

    Args:
        data (bytes): pickle data to decode
        **kwargs: same as in 'pickle.loads'

    Returns:
        pickle data decoded
    """
    return pickle.loads(data, **kwargs)


def encode_pickle(data, **kwargs):
    """
    Encode pickle data to bytes
    Backend used: pickle.dumps (pickle package)

    Args:
        data: pickle data to encode
        **kwargs: same as in 'pickle.dumps'

    Returns:
        bytes: pickle data encoded
    """
    return pickle.dumps(data, **kwargs)


def read_pickle(openfile, **kwargs):
    """
    Read a open pickle file
    Backend used: pickle.loads (pickle package)

    Args:
        openfile (file-like): file to read, must have been opened
        **kwargs: same as in 'pickle.loads'

    Returns:
        pickle data read from the file provided
    """
    # return pickle.load(openfile, **kwargs) # does not work when combined with zstd
    data = openfile.read()
    return pickle.loads(data, **kwargs)


def write_pickle(openfile, data, **kwargs):
    """
    Write in a open pickle file
    Backend used: pickle.dumps (pickle package)

    Args:
        openfile (file-like): file to write, must have been opened
        data: pickle data to save
        **kwargs: same as in 'pickle.dump'
    """
    return pickle.dump(data, openfile, **kwargs)


def load_pickle(filename, fs=None, **kwargs):
    """
    Load a pickle file
    Backend used: pickle.loads (pickle package)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        **kwargs: same as in 'pickle.loads'

    Returns:
        pickle data loaded from the file provided
    """
    with _generic_open(filename, 'rb', fs=fs) as f:
        data = read_pickle(f, **kwargs)
    return data


def save_pickle(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a pickle file
    Backend used: pickle.dumps (pickle package)

    Args:
        filename (str): file name to save data to
        data: pickle data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'pickle.dump'
    """
    with _generic_open(filename, 'wb', fs=fs, makedirs=makedirs) as f:
        write_pickle(f, data, **kwargs)


def help_pickle():
    """
    Print help for pickle io
    """
    _help_default("pickle")




import importlib as _importlib


__all__ = [
    'decode_avi',
    'encode_avi',
    'read_avi',
    'write_avi',
    'load_avi',
    'save_avi',
    'help_avi',
    'decode_bytes',
    'encode_bytes',
    'read_bytes',
    'write_bytes',
    'load_bytes',
    'save_bytes',
    'help_bytes',
    'decode_csv',
    'encode_csv',
    'read_csv',
    'write_csv',
    'load_csv',
    'save_csv',
    'help_csv',
    'decode_gif',
    'encode_gif',
    'read_gif',
    'write_gif',
    'load_gif',
    'save_gif',
    'help_gif',
    'decode_gz',
    'encode_gz',
    'read_gz',
    'write_gz',
    'load_gz',
    'save_gz',
    'help_gz',
    'decode_image',
    'encode_image',
    'read_image',
    'write_image',
    'load_image',
    'save_image',
    'help_image',
    'decode_jpg',
    'encode_jpg',
    'read_jpg',
    'write_jpg',
    'load_jpg',
    'save_jpg',
    'help_jpg',
    'decode_json',
    'encode_json',
    'read_json',
    'write_json',
    'load_json',
    'save_json',
    'help_json',
    'decode_mp4',
    'encode_mp4',
    'read_mp4',
    'write_mp4',
    'load_mp4',
    'save_mp4',
    'help_mp4',
    'decode_npy',
    'encode_npy',
    'read_npy',
    'write_npy',
    'load_npy',
    'save_npy',
    'help_npy',
    'decode_pickle',
    'encode_pickle',
    'read_pickle',
    'write_pickle',
    'load_pickle',
    'save_pickle',
    'help_pickle',
    'decode_png',
    'encode_png',
    'read_png',
    'write_png',
    'load_png',
    'save_png',
    'help_png',
    'decode_tar',
    'encode_tar',
    'read_tar',
    'write_tar',
    'load_tar',
    'save_tar',
    'help_tar',
    'decode_txt',
    'encode_txt',
    'read_txt',
    'write_txt',
    'load_txt',
    'save_txt',
    'help_txt',
    'decode_video',
    'encode_video',
    'read_video',
    'write_video',
    'load_video',
    'save_video',
    'help_video',
    'decode_xml',
    'encode_xml',
    'read_xml',
    'write_xml',
    'load_xml',
    'save_xml',
    'help_xml',
    'decode_yaml',
    'encode_yaml',
    'read_yaml',
    'write_yaml',
    'load_yaml',
    'save_yaml',
    'help_yaml',
    'decode_zip',
    'encode_zip',
    'read_zip',
    'write_zip',
    'load_zip',
    'save_zip',
    'help_zip',
    'decode_zst',
    'encode_zst',
    'read_zst',
    'write_zst',
    'load_zst',
    'save_zst',
    'help_zst',
]


def __getattr__(attr):
    if attr in {
        'decode_avi',
        'encode_avi',
        'read_avi',
        'write_avi',
        'load_avi',
        'save_avi',
        'help_avi',
    }:
        aviio = _importlib.import_module('iotools.aviio')
        return getattr(aviio, attr)
    if attr in {
        'decode_bytes',
        'encode_bytes',
        'read_bytes',
        'write_bytes',
        'load_bytes',
        'save_bytes',
        'help_bytes',
    }:
        bytesio = _importlib.import_module('iotools.bytesio')
        return getattr(bytesio, attr)
    if attr in {
        'decode_csv',
        'encode_csv',
        'read_csv',
        'write_csv',
        'load_csv',
        'save_csv',
        'help_csv',
    }:
        csvio = _importlib.import_module('iotools.csvio')
        return getattr(csvio, attr)
    if attr in {
        'decode_gif',
        'encode_gif',
        'read_gif',
        'write_gif',
        'load_gif',
        'save_gif',
        'help_gif',
    }:
        gifio = _importlib.import_module('iotools.gifio')
        return getattr(gifio, attr)
    if attr in {
        'decode_gz',
        'encode_gz',
        'read_gz',
        'write_gz',
        'load_gz',
        'save_gz',
        'help_gz',
    }:
        gzio = _importlib.import_module('iotools.gzio')
        return getattr(gzio, attr)
    if attr in {
        'decode_image',
        'encode_image',
        'read_image',
        'write_image',
        'load_image',
        'save_image',
        'help_image',
    }:
        imageio = _importlib.import_module('iotools.imageio')
        return getattr(imageio, attr)
    if attr in {
        'decode_jpg',
        'encode_jpg',
        'read_jpg',
        'write_jpg',
        'load_jpg',
        'save_jpg',
        'help_jpg',
    }:
        jpgio = _importlib.import_module('iotools.jpgio')
        return getattr(jpgio, attr)
    if attr in {
        'decode_json',
        'encode_json',
        'read_json',
        'write_json',
        'load_json',
        'save_json',
        'help_json',
    }:
        jsonio = _importlib.import_module('iotools.jsonio')
        return getattr(jsonio, attr)
    if attr in {
        'decode_mp4',
        'encode_mp4',
        'read_mp4',
        'write_mp4',
        'load_mp4',
        'save_mp4',
        'help_mp4',
    }:
        mp4io = _importlib.import_module('iotools.mp4io')
        return getattr(mp4io, attr)
    if attr in {
        'decode_npy',
        'encode_npy',
        'read_npy',
        'write_npy',
        'load_npy',
        'save_npy',
        'help_npy',
    }:
        npyio = _importlib.import_module('iotools.npyio')
        return getattr(npyio, attr)
    if attr in {
        'decode_pickle',
        'encode_pickle',
        'read_pickle',
        'write_pickle',
        'load_pickle',
        'save_pickle',
        'help_pickle',
    }:
        pickleio = _importlib.import_module('iotools.pickleio')
        return getattr(pickleio, attr)
    if attr in {
        'decode_png',
        'encode_png',
        'read_png',
        'write_png',
        'load_png',
        'save_png',
        'help_png',
    }:
        pngio = _importlib.import_module('iotools.pngio')
        return getattr(pngio, attr)
    if attr in {
        'decode_tar',
        'encode_tar',
        'read_tar',
        'write_tar',
        'load_tar',
        'save_tar',
        'help_tar',
    }:
        tario = _importlib.import_module('iotools.tario')
        return getattr(tario, attr)
    if attr in {
        'decode_txt',
        'encode_txt',
        'read_txt',
        'write_txt',
        'load_txt',
        'save_txt',
        'help_txt',
    }:
        txtio = _importlib.import_module('iotools.txtio')
        return getattr(txtio, attr)
    if attr in {
        'decode_video',
        'encode_video',
        'read_video',
        'write_video',
        'load_video',
        'save_video',
        'help_video',
    }:
        videoio = _importlib.import_module('iotools.videoio')
        return getattr(videoio, attr)
    if attr in {
        'decode_xml',
        'encode_xml',
        'read_xml',
        'write_xml',
        'load_xml',
        'save_xml',
        'help_xml',
    }:
        xmlio = _importlib.import_module('iotools.xmlio')
        return getattr(xmlio, attr)
    if attr in {
        'decode_yaml',
        'encode_yaml',
        'read_yaml',
        'write_yaml',
        'load_yaml',
        'save_yaml',
        'help_yaml',
    }:
        yamlio = _importlib.import_module('iotools.yamlio')
        return getattr(yamlio, attr)
    if attr in {
        'decode_zip',
        'encode_zip',
        'read_zip',
        'write_zip',
        'load_zip',
        'save_zip',
        'help_zip',
    }:
        zipio = _importlib.import_module('iotools.zipio')
        return getattr(zipio, attr)
    if attr in {
        'decode_zst',
        'encode_zst',
        'read_zst',
        'write_zst',
        'load_zst',
        'save_zst',
        'help_zst',
    }:
        zstio = _importlib.import_module('iotools.zstio')
        return getattr(zstio, attr)
    raise AttributeError(attr)



"""
This module provides io for gzip files (.gz extension)
See iotools.gzio.help_gz() for more info
"""

import gzip

from iotools._help_utils import _help_default
from iotools._generic_open import _generic_open


def decode_gz(data, **kwargs):
    """
    Decode gz data from str
    Backend used: gzip.decompress (gzip package)

    Args:
        data (bytes): gzip data to decode
        **kwargs: same as in 'gzip.decompress'

    Returns:
        bytes: gzip data decoded
    """
    return gzip.decompress(data, **kwargs)


def encode_gz(data, **kwargs):
    """
    Encode gz data to str
    Backend used: gzip.compress (gzip package)

    Args:
        data (bytes | str): gzip data to encode
        **kwargs: same as in 'gzip.compress'

    Returns:
        bytes: gzip data encoded
    """
    if isinstance(data, str):
        data = data.encode("utf-8")
    return gzip.compress(data, **kwargs)


def read_gz(openfile, **kwargs):
    """
    Read a open gz file
    Backend used: gzip.decompress (gzip package)

    Args:
        openfile (file-like): file to read, must have been opened
        **kwargs: same as in 'gzip.decompress'

    Returns:
        bytes: gzip data read from the file provided
    """
    return gzip.decompress(openfile.read(), **kwargs)


def write_gz(openfile, data, **kwargs):
    """
    Write in a open gz file
    Backend used: gzip.compress (gzip package)

    Args:
        openfile (file-like): file to write, must have been opened
        data (bytes | str): gzip data to save
        **kwargs: same as in 'gzip.compress'
    """
    if isinstance(data, str):
        data = data.encode("utf-8")
    openfile.write(gzip.compress(data, **kwargs))


def load_gz(filename, fs=None, **kwargs):
    """
    Load a gz file
    Backend used: gzip.decompress (gzip package)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        **kwargs: same as in 'gzip.decompress'

    Returns:
        bytes: loaded gzip data
    """
    with _generic_open(filename, 'rb', fs=fs) as f:
        data = read_gz(f, **kwargs)
    return data


def save_gz(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a gz file
    Backend used: gzip.compress (gzip package)

    Args:
        filename (str): file name to save data to
        data (bytes | str): gzip data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'gzip.compress'
    """
    with _generic_open(filename, 'wb', fs=fs, makedirs=makedirs) as f:
        write_gz(f, data, **kwargs)


def help_gz():
    """
    Print help for gzip io
    """
    _help_default("gz")



"""
This module provides io for tar files (.tar extension)
See iotools.tario.help_tar() for more info
"""

import tarfile
from io import BytesIO

from iotools._help_utils import _help_default
from iotools._generic_open import _generic_open


def decode_tar(data, **kwargs):
    """
    Decode tar data from bytes
    Backends used: tarfile.open, tarfile.getnames, tarfile.extractfile (tarfile package)

    Args:
        data (bytes): tar data to decode
        **kwargs: same as in 'tarfile.open'

    Returns:
        dict{filename: bytes}: tar data decoded
    """
    b = BytesIO(data)
    t = tarfile.open(fileobj=b, **kwargs)
    ret = {}
    for subfilename in t.getnames():
        subdata = t.extractfile(subfilename)
        if subdata is not None:
            ret[subfilename] = subdata.read()
    t.close()
    b.close()
    return ret


def encode_tar(data, **kwargs):
    """
    Encode tar data to bytes
    Backends used: tarfile.open, tarfile.TarInfo, tarfile.addfile (tarfile package)

    Args:
        data (dict{filename: bytes}): tar data to encode
        **kwargs: same as in 'tarfile.open'

    Returns:
        bytes: tar data encoded
    """
    b = BytesIO()
    t = tarfile.open(fileobj=b, mode="w", **kwargs)
    for subfilename, raw_data in data.items():
        ti = tarfile.TarInfo(subfilename)
        ti.size = len(raw_data)
        if isinstance(raw_data, str):
            raw_data = raw_data.encode('utf-8')
        t.addfile(ti, BytesIO(raw_data))
    b.seek(0)
    ret = b.read()
    t.close()
    b.close()
    return ret


def read_tar(openfile, **kwargs):
    """
    Read a open tar file
    Backends used: tarfile.open, tarfile.getnames, tarfile.extractfile (tarfile package)

    Args:
        openfile (file-like): file to read, must have been opened
        **kwargs: same as in 'tarfile.open'

    Returns:
        dict{filename: bytes}: tar data read from the file provided
    """
    return decode_tar(openfile.read(), **kwargs)


def write_tar(openfile, data, **kwargs):
    """
    Write in a open tar file
    Backends used: tarfile.open, tarfile.TarInfo, tarfile.addfile (tarfile package)

    Args:
        openfile (file-like): file to write, must have been opened
        data (dict{filename: bytes}): tar data to save
        **kwargs: same as in 'tarfile.open'
    """
    openfile.write(encode_tar(data, **kwargs))


def load_tar(filename, fs=None, **kwargs):
    """
    Load a tar file
    Backends used: tarfile.open, tarfile.getnames, tarfile.extractfile (tarfile package)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        **kwargs: same as in 'tarfile.open'

    Returns:
        dict{filename: bytes}: tar data loaded from the file provided
    """
    with _generic_open(filename, 'rb', fs=fs) as f:
        data = read_tar(f, **kwargs)
    return data


def save_tar(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a tar file
    Backends used: tarfile.open, tarfile.TarInfo, tarfile.addfile (tarfile package)

    Args:
        filename (str): file name to save data to
        data (dict{filename: bytes}): tar data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'tarfile.open'
    """
    with _generic_open(filename, 'wb', fs=fs, makedirs=makedirs) as f:
        write_tar(f, data, **kwargs)


def help_tar():
    """
    Print help for tar io
    """
    _help_default("tar")



"""
This module provides io for binary text files
See iotools.bytesio.help_bytes() for more info
"""

import os

from iotools._help_utils import _help_default
from iotools._generic_open import _generic_open


def decode_bytes(data, **kwargs):
    """
    Decode bytes data, which is the identity function.

    Args:
        data (str | bytes): bytes data to decode
        **kwargs: same as in 'str.encode'

    Returns:
        bytes: decoded data
    """
    if isinstance(data, str):
        return data.encode(**kwargs)
    return data


def encode_bytes(data, **kwargs):
    """
    Encode bytes data, which is the identity function.

    Args:
        data (str | bytes): bytes data to encode
        **kwargs: same as in 'str.encode'

    Returns:
        bytes: encoded data
    """
    if isinstance(data, str):
        return data.encode(**kwargs)
    return data


def read_bytes(openfile, **kwargs):
    """
    Read an open binary text file
    Backend used: TextIOWrapper.read (builtin)

    Args:
        openfile (file-like): file to read, must have been opened
        **kwargs: same as in 'TextIOWrapper.read'

    Returns:
        bytes: data read from the file provided
    """
    data = openfile.read(**kwargs)
    if (os.name == "nt"):  # On windows
        data = data.replace(b'\r\n', b'\n')  # Windows has different newlines encoding
    return data


def write_bytes(openfile, data, **kwargs):
    """
    Write binary text file
    Backend used: TextIOWrapper.write (builtin)

    Args:
        openfile (file-like): file to write, must have been opened
        data (bytes): text data to save
        **kwargs: same as in 'TextIOWrapper.write'
    """
    openfile.write(data, **kwargs)


def load_bytes(filename, fs=None, **kwargs):
    """
    Load a binary text file
    Backend used: TextIOWrapper.read (builtin)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        **kwargs: same as in 'TextIOWrapper.read'

    Returns:
        bytes: loaded data
    """
    with _generic_open(filename, "rb", fs=fs) as f:
        data = read_bytes(f, **kwargs)
    return data


def save_bytes(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a binary text file
    Backend used: TextIOWrapper.write (builtin)

    Args:
        filename (str): file name to save data to
        data (bytes): text data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'TextIOWrapper.write'
    """
    with _generic_open(filename, "wb", fs=fs, makedirs=makedirs) as f:
        write_bytes(f, data, **kwargs)


def help_bytes():
    """
    Print help for bytes io
    """
    _help_default("bytes")



"""
This module provides io for numpy files (.npy extension)
See iotools.npyio.help_npy() for more info
"""

from iotools._help_utils import _help_default
from iotools._generic_open import _generic_open
from iotools._default_decode_encode import _default_decode_bytes, _default_encode_bytes

try:
    import numpy as np
except ModuleNotFoundError:
    from iotools.settings import settings
    from iotools._missing_module_helper import _EmptyModule, _error_msg_one
    if settings["SHOW_IMPORT_WARNINGS"]:
        print("Warning:", _error_msg_one("iotools.npyio", "numpy"))
    np = _EmptyModule("numpy")


def decode_npy(data, **kwargs):
    """
    Decode npy data from bytes
    Backend used: numpy.load (numpy package)

    Args:
        data (bytes): numpy data to decode
        **kwargs: same as in 'numpy.load'

    Returns:
        numpy.ndarray: numpy data decoded
    """
    return _default_decode_bytes(np.load, data, **kwargs)


def encode_npy(data, **kwargs):
    """
    Encode npy data to bytes
    Backend used: numpy.save (numpy package)

    Args:
        data (numpy.ndarray): numpy data to encode
        **kwargs: same as in 'numpy.save'

    Returns:
        bytes: numpy data encoded
    """
    return _default_encode_bytes(np.save, data, **kwargs)


def read_npy(openfile, **kwargs):
    """
    Read a open npy file
    Backend used: numpy.load (numpy package)

    Args:
        openfile (file-like): file to read, must have been opened
        **kwargs: same as in 'numpy.load'

    Returns:
        numpy.ndarray: numpy data read from the file provided
    """
    return np.load(openfile, **kwargs)


def write_npy(openfile, data, **kwargs):
    """
    Write in a open npy file
    Backend used: numpy.save (numpy package)

    Args:
        openfile (file-like): file to write, must have been opened
        data (numpy.ndarray): numpy data to save
        **kwargs: same as in 'numpy.save'
    """
    np.save(openfile, data, **kwargs)


def load_npy(filename, fs=None, **kwargs):
    """
    Load a npy file
    Backend used: numpy.load (numpy package)

    Args:
        filename (str): file name to load
        fs: file system to use. Defaults to None (local filesystem)
        **kwargs: same as in 'numpy.load'

    Returns:
        numpy.ndarray: loaded numpy data
    """
    with _generic_open(filename, 'rb', fs=fs) as f:
        data = read_npy(f, **kwargs)
    return data


def save_npy(filename, data, fs=None, makedirs=False, **kwargs):
    """
    Save a npy file
    Backend used: numpy.save (numpy package)

    Args:
        filename (str): file name to save data to
        data (numpy.ndarray): numpy data to save
        fs: file system to use. Defaults to None (local filesystem)
        makedirs (bool): if True, creates directory of filename (default is False)
        **kwargs: same as in 'numpy.save'
    """
    with _generic_open(filename, 'wb', fs=fs, makedirs=makedirs) as f:
        write_npy(f, data, **kwargs)


def help_npy():
    """
    Print help for npy io
    """
    _help_default("npy")
